# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class DictToJson(GeneralLogicNode):
    NICE_NAME = "Dict to json"
    HELP = "Write dict to JSON file"

    INPUTS_DICT = {
        "in_dict": {"type": dict},
        "json_filepath_to_write": {"type": str},
    }

    def run(self):
        import json

        in_dict = self.get_attribute_value("in_dict")
        json_filepath_to_write = self.get_attribute_value("json_filepath_to_write")

        with open(json_filepath_to_write, "w") as f:
            json.dump(in_dict, f, indent=4)


class DictToYaml(GeneralLogicNode):
    NICE_NAME = "Dict to yaml"
    HELP = "Write dict to YAML file"

    INPUTS_DICT = {
        "in_dict": {"type": dict},
        "yaml_filepath_to_write": {"type": str},
    }

    def run(self):
        import yaml

        in_dict = self.get_attribute_value("in_dict")
        path = self.get_attribute_value("yaml_filepath_to_write")
        with open(path, "w") as file:
            yaml.dump(in_dict, file, indent=4)


class ListToTxt(GeneralLogicNode):
    INPUTS_DICT = {
        "in_list": {"type": list},
        "txt_filepath_to_write": {"type": str},
    }

    def run(self):
        in_list = self.get_attribute_value("in_list")
        path = self.get_attribute_value("txt_filepath_to_write")
        with open(path, "w") as file:
            for elem in in_list:
                file.write(str(elem) + "\n")


class TableToCsv(GeneralLogicNode):
    INPUTS_DICT = {
        "in_table": {"type": list},
        "csv_filepath_to_write": {"type": str},
    }

    def run(self):
        import csv

        in_table = self.get_attribute_value("in_table")
        csv_filepath_to_write = self.get_attribute_value("csv_filepath_to_write")
        with open(csv_filepath_to_write, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(in_table)


class CreateTempFile(GeneralLogicNode):
    NICE_NAME = "Create temp file"

    INPUTS_DICT = {"suffix": {"type": str, "optional": True}}
    OUTPUTS_DICT = {"tempfile_path": {"type": str}}

    def run(self):
        import tempfile

        out_tempfile = ""

        suffix = self.get_attribute_value("suffix")
        if suffix:
            temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
            temp_file.close()
            out_tempfile = temp_file.name
        else:
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            temp_file.close()
            out_tempfile = temp_file.name

        self.set_output("tempfile_path", out_tempfile)
        LOGGER.debug("Created tempfile: {}".format(out_tempfile))



# -*- coding: UTF-8 -*-
from __future__ import annotations

__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import datetime
import getpass
import os
import yaml

from PySide2 import QtCore

from all_nodes import constants
from all_nodes import utils
from all_nodes.analytics import analytics
from all_nodes.graphic.widgets.global_signaler import GlobalSignaler

from all_nodes.logic import class_registry
from all_nodes.logic.class_registry import CLASS_REGISTRY as CR
from all_nodes.logic.logic_node import GeneralLogicNode


GS = GlobalSignaler()

LOGGER = utils.get_logger(__name__)


# -------------------------------- LOGIC SCENE -------------------------------- #
class LogicScene:
    def __init__(self):
        self.scene_name = None

        self.context = None

        self.all_logic_nodes = set()
        self.class_counter = dict()

        self.pasted_count = 0

        self.thread_manager = QtCore.QThreadPool.globalInstance()

        LOGGER.debug("Initialized logic scene")

    # NODE ADDITION AND DELETION ----------------------
    def add_node_by_name(
        self, node_classname: str, rename_on_create=True
    ) -> GeneralLogicNode:
        """
        Add a new node to the logic scene based on the given node class name.

        Args:
            node_classname (str): The name of the class of the node to be added.
            rename_on_create (bool, optional): Whether to rename the node if it already exists. Defaults to True.

        Returns:
            GeneralLogicNode: The newly created node.

        Raises:
            LogicSceneError: If no class with the given name is found.
        """
        all_classes = CR.get_all_classes()

        for lib in sorted(all_classes):
            for m in all_classes[lib]:
                for name, cls in all_classes[lib][m]["classes"]:
                    if node_classname == name:
                        new_logic_node = cls()
                        self.all_logic_nodes.add(new_logic_node)
                        if node_classname not in self.class_counter:
                            self.class_counter[node_classname] = 1
                        else:
                            self.class_counter[node_classname] += 1
                            if rename_on_create:
                                self.rename_node(
                                    new_logic_node,
                                    node_classname
                                    + "_"
                                    + str(self.class_counter[node_classname]),
                                )

                        if self.context:
                            new_logic_node.set_context(self.context)
                        return new_logic_node

        raise LogicSceneError("No class {} was found!".format(node_classname))

    def clear(self):
        """
        Clear the logic scene, removing all logic nodes
        """
        self.all_logic_nodes = set()
        LOGGER.info("Cleared logic scene")

    def remove_node_by_name(self, node_fullname):
        for node in self.all_logic_nodes:
            if node.node_name == node_fullname:
                self.all_logic_nodes.remove(node)
                LOGGER.info("Removed logic node {}".format(node_fullname))
                return
        raise RuntimeError("No node matches name" + node_fullname)

    # NODE RETRIEVAL ----------------------
    def all_nodes(self):
        return self.all_logic_nodes

    def node_count(self):
        return len(self.all_logic_nodes)

    def get_starting_nodes(self):
        starting_nodes = []
        for n in self.all_logic_nodes:
            if n.is_starting_node():
                LOGGER.info("Node {} to be used as starting node".format(n.node_name))
                starting_nodes.append(n)

        if not starting_nodes:
            LOGGER.warning("No starting nodes found in this logic scene")

        return starting_nodes

    def to_node(self, node_full_name):
        for node in self.all_logic_nodes:
            if node.node_name == node_full_name:
                return node

    def to_attr(self, attr_full_name):
        for node in self.all_logic_nodes:
            for attr in node.all_attributes:
                if attr.full_name == attr_full_name:
                    return attr

    # NODE MANIPULATION ----------------------
    def rename_node(self, node, new_name: str) -> bool:
        for n in self.all_logic_nodes:
            if n.uuid == node.uuid:
                continue
            if n.node_name == new_name:
                LOGGER.error("Node {} already exists!".format(new_name))
                return False

        return node.rename(new_name)

    def rename_node_with_namespace(self, node, new_name: str):
        """
        Rename a node with a given new name, while ensuring that the new name is unique within the scene.

        Args:
            node (LogicNode): The node to be renamed.
            new_name (str): The new name for the node.

        Raises:
            LogicSceneError: If a node with the same name already exists in the scene.

        Returns:
            bool: True if the node could be renamed
        """
        for n in self.all_logic_nodes:
            if n.uuid == node.uuid:
                continue
            if n.node_name == new_name:
                raise LogicSceneError("Node {} already exists!".format(new_name))

        return node.force_rename(new_name)

    def set_context_to_nodes(self):
        """
        Set the context of each node in the logic scene to the context that this scene is in.
        """
        for node in self.all_logic_nodes:
            node.set_context(self.context)

    def connect_attrs_by_name(self, source_attr_name: str, target_attr_name: str):
        """
        Connect two attributes in the logic scene by their names.

        Args:
            source_attr_name (str): The dot-separated name of the source attribute.
            target_attr_name (str): The dot-separated name of the target attribute.

        Raises:
            LogicSceneError: If any of the following conditions are met:
                - The source node does not exist.
                - The target node does not exist.
                - The source attribute does not exist.
                - The target attribute does not exist.
        """
        source_attr, target_attr = None, None
        for node in self.all_logic_nodes:
            for attr in node.all_attributes:
                if attr.dot_name == source_attr_name:
                    source_attr = attr
                    break
                elif attr.dot_name == target_attr_name:
                    target_attr = attr
                    break
            if source_attr and target_attr:
                source_attr.connect_to_other(target_attr)  # TODO check for result
                return

        # Errors
        source_node_name = source_attr_name.rsplit(".", 1)[0]
        target_node_name = target_attr_name.rsplit(".", 1)[0]
        if self.to_node(source_node_name) is None:
            raise LogicSceneError(
                "Cannot connect! {} -> {}, source node '{}' does not exist".format(
                    source_attr_name, target_attr_name, source_node_name
                )
            )
        elif self.to_node(target_node_name) is None:
            raise LogicSceneError(
                "Cannot connect! {} -> {}, target node '{}' does not exist".format(
                    source_attr_name, target_attr_name, target_node_name
                )
            )
        elif source_attr is None:
            raise LogicSceneError(
                "Cannot connect! {} -> {}, source attribute '{}' does not exist".format(
                    source_attr_name, target_attr_name, source_attr_name
                )
            )
        elif target_attr is None:
            raise LogicSceneError(
                "Cannot connect! {} -> {}, target attribute '{}' does not exist".format(
                    source_attr_name, target_attr_name, target_attr_name
                )
            )

    # SAVE AND LOAD ----------------------
    def convert_scene_to_dict(self):
        """
        Convert the scene to a dict (that can be then written out as json, yaml...)

        Returns:
            dict: dict that represents the scene
        """
        scene_dict = dict()

        all_node_names = sorted([n.node_name for n in self.all_logic_nodes])
        if all_node_names:
            scene_dict["nodes"] = list()
            for name in all_node_names:
                for node in self.all_logic_nodes:
                    if node.node_name == name:
                        node_dict = node.get_node_basic_dict()
                        scene_dict["nodes"].append(node_dict)
                        break

        connections = set()
        for node in self.all_logic_nodes:
            out_conns = node.get_out_connections()
            if out_conns:
                for c in out_conns:
                    connections.add(" -> ".join(c))
        if connections:
            scene_dict["connections"] = sorted(list(connections))

        return scene_dict

    def save_to_file(self, filepath: str, scene_dict: dict = None) -> None:
        """
        Save the scene out

        Args:
            filepath (str): filepath to save to
            scene_dict (dict, optional): scene info to write out. Defaults to None.
        """
        # Get scene data
        if scene_dict is None:
            scene_dict = self.convert_scene_to_dict()

        # Save type
        file_type = "scene"
        _, ext = os.path.splitext(filepath)
        if ext == ".ctx":
            file_type = "context"

        save_type = "created"
        if os.path.exists(filepath):
            save_type = "modified"

        # Actual save
        with open(filepath, "w") as file:
            header = "# {} {}".format(
                file_type.upper(), os.path.splitext(os.path.basename(filepath))[0]
            )
            file.write(header)
            file.write("\n# " + "-" * (len(header) - 2))
            file.write("\n# Description: \n")

            if "nodes" in scene_dict:
                file.write(
                    "\n# Nodes section: overall list of nodes to be created\nnodes:\n"
                )
                yaml.dump(scene_dict["nodes"], file, sort_keys=True)

            if "connections" in scene_dict:
                file.write(
                    "\n# Connections section: connections to be done between nodes\n"
                    "connections:\n"
                )
                yaml.dump(scene_dict["connections"], file)

            if "annotations" in scene_dict:
                file.write(
                    "\n# Annotations section: list of annotations in the scene\n"
                    "annotations:\n"
                )
                yaml.dump(scene_dict["annotations"], file, sort_keys=True)

            file.write(
                f"\n\n# {file_type.capitalize()} {save_type} at: {datetime.datetime.now()}"
            )
            file.write(f"\n# {save_type.capitalize()} by: {getpass.getuser()}")

        LOGGER.info("Wrote scene to file: {}".format(filepath))

    def load_from_file(self, scene_path: str, namespace: str = None) -> list:
        """
        Load scene from a given file

        Args:
            scene_path (str): scene filepath or alias to load
            namespace (str, optional): namespace to apply to the created nodes while loading scene. Defaults to None.

        Raises:
            LogicSceneError: if the scene requested doesnt exist, or is malformed

        Returns:
            list: of newly created nodes
        """
        # See if we need namespace
        if self.all_nodes():
            if namespace is None:
                namespace = self.get_namespace()
            else:
                namespace += "::"
        else:
            namespace = ""

        if namespace:
            LOGGER.debug("Using namespace: {}".format(namespace))

        # Alias
        all_scenes = CR.get_all_scenes()  # TODO move this into the class
        if not os.path.isfile(scene_path):
            LOGGER.info(
                "Cannot find scene with path '{}', trying to find it as alias".format(
                    scene_path
                )
            )
            found_scene_path = class_registry.get_scene_from_alias(
                all_scenes, scene_path
            )
            if not found_scene_path:
                raise LogicSceneError(
                    "Cannot find scene for alias '{}'".format(scene_path)
                )
            else:
                scene_path = found_scene_path

        # Filepath
        if self.context:
            utils.print_separator("Loading context " + scene_path)
        else:
            utils.print_separator("Loading scene " + scene_path)
        scene_dict = dict()
        with open(scene_path, "r") as file:
            scene_dict = yaml.safe_load(file)

        if not scene_dict:
            return  # TODO raise error

        # Create nodes
        new_nodes = []
        for node in scene_dict.get("nodes", []):
            node_name = next(iter(node))
            n = self.add_node_by_name(node[node_name]["class_name"], False)
            new_nodes.append(n)
            if namespace:
                if not self.rename_node_with_namespace(n, namespace + node_name):
                    raise LogicSceneError(
                        "Node cannot be renamed to: {}".format(namespace + node_name)
                    )
            else:
                if not self.rename_node(n, node_name):
                    raise LogicSceneError(
                        "Node cannot be renamed to: {}".format(namespace + node_name)
                    )
            attrs_to_set = node[node_name].get("node_attributes")
            if attrs_to_set:
                for attr in attrs_to_set:
                    n.set_attribute_value(attr, attrs_to_set[attr])

            active = node[node_name].get("active", True)
            if not active:
                n.toggle_activated()

        # Create connections
        for connection in scene_dict.get("connections", []):
            attrs_to_connect = connection.split("->")
            source_attr_name, target_attr_name = (
                namespace + attrs_to_connect[0].strip(),
                namespace + attrs_to_connect[1].strip(),
            )
            self.connect_attrs_by_name(source_attr_name, target_attr_name)

        return new_nodes

    # SCENE PROPERTIES ----------------------
    def set_name(self, new_name: str):
        """
        Set the name of the scene.

        Parameters:
            new_name (str): The new name to set for the scene.
        """
        self.scene_name = new_name

    def get_namespace(self):
        """
        Get a generic namespace for pasted nodes.

        Returns:
            str: The namespace string.
        """
        self.pasted_count += 1
        return "pasted_{}::".format(self.pasted_count)

    # RESET NODES ----------------------
    def reset_all_nodes(self):
        """
        Reset all logic nodes of the scene.
        """
        LOGGER.debug("Resetting all logic nodes of scene {}".format(self.scene_name))
        for node in self.all_logic_nodes:
            node.reset()

    def soft_reset_all_nodes(self):
        """
        Soft-reset all logic nodes of the scene.
        """
        LOGGER.debug(
            "Soft-resetting all logic nodes of scene {}".format(self.scene_name)
        )
        for node in self.all_logic_nodes:
            node.soft_reset()

    # EXECUTION ----------------------
    def run_all_nodes(self, spawn_thread=True):
        """
        Run all nodes in the scene.

        Parameters:
            spawn_thread (bool, optional): Whether to spawn a new thread to run the nodes in . Defaults to True.
        """
        if spawn_thread:
            worker = Worker(self._run_all_nodes)
            self.thread_manager.start(worker)
        else:
            self._run_all_nodes()

    def run_all_nodes_batch(self):
        """For non-GUI, we cannot spawn threads"""
        # TODO investigate a better way
        self._run_all_nodes()

    def run_list_of_nodes(self, nodes_to_execute: list, spawn_thread: bool = True):
        """
        Execute a list of nodes.

        Parameters:
            nodes_to_execute (list): A list of nodes to be executed.
            spawn_thread (bool, optional): Whether to spawn a new thread to run the nodes in. Defaults to True.
        """
        if spawn_thread:
            worker = Worker(self._run_list_of_nodes, nodes_to_execute)
            self.thread_manager.start(worker)
        else:
            self._run_list_of_nodes(nodes_to_execute)

    def _run_all_nodes(self):
        """
        Execute all the nodes in this logic scene.
        """
        # Feedback
        if self.scene_name:
            utils.print_separator("Running {}".format(self.scene_name))
        else:
            utils.print_separator("Running logic scene")

        # Execution
        for node in self.get_starting_nodes():
            node._run()
        LOGGER.info("Finished running logic scene")

        # Mark nodes that were skipped
        for node in self.all_nodes():
            if node.success == constants.NOT_RUN:
                node.mark_skipped()

        # Analytics
        # TODO separate into own method in case analytics crash
        LOGGER.info("Gathering analytics")
        node_properties_list = []
        for node in self.all_logic_nodes:
            node_properties_list.append(node.get_node_full_dict())
            if node.IS_CONTEXT:
                for i_node in (
                    node.internal_scene.all_logic_nodes
                ):  # TODO make this properly recursive
                    node_properties_list.append(i_node.get_node_full_dict())
        analytics.submit_bulk_analytics(node_properties_list)

    def _run_list_of_nodes(self, nodes_to_execute: list):
        """
        Run a list of nodes.

        Parameters:
            nodes_to_execute (list): A list of nodes to be executed.
        """
        for node in nodes_to_execute:
            node.run_single()

    # FEEDBACK GATHERING ----------------------
    # TODO make these properly recursive, mark contexts appropriately
    def gather_failed_nodes_logs(self) -> list:
        """
        Gather all failed logs from the logic nodes in the scene.

        Returns:
            list: A list of failed logs.
        """
        failed_log = []
        for node in self.all_logic_nodes:
            if node.success in [constants.FAILED, constants.ERROR]:
                for line in node.fail_log:
                    failed_log.append(node.full_name + ": " + line)
        return failed_log

    def gather_errored_nodes_logs(self) -> list:
        """
        Gather the error logs from all nodes in the scene.

        Returns:
            list: A list of error logs.
        """
        errored_log = []
        for node in self.all_logic_nodes:
            if node.success in [constants.FAILED, constants.ERROR]:
                for line in node.error_log:
                    errored_log.append(node.full_name + ": " + line)
        return errored_log


# -------------------------------- UTILITY -------------------------------- #
class LogicSceneError(Exception):
    pass


# -------------------------------- WORKER -------------------------------- #
class Worker(QtCore.QRunnable):
    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """
        Initialise the runner function with passed args, kwargs.
        """
        self.fn(*self.args, **self.kwargs)
        GS.signals.execution_finished.emit()
        GS.signals.attribute_editor_global_refresh_requested.emit()



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

from all_nodes import constants
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes.logic.class_registry import CLASS_REGISTRY as CR


from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class NodePanel(QtWidgets.QWidget):
    """Widget representation of a logic node and its attributes"""

    def __init__(self, logic_node: GeneralLogicNode):
        QtWidgets.QWidget.__init__(self)

        self.grid_layout = QtWidgets.QGridLayout()
        self.setLayout(self.grid_layout)

        self.logic_node = logic_node

        self.color = self.get_node_color()
        self.get_node_color()

        self.row_count = 0
        self.represent_node()

    def get_node_color(self):
        """
        Examine the configs and find the color associated to the represented node

        Returns:
            QtGui.QColor: color that corresponds to the logic node

        """
        all_classes = CR.get_all_classes()
        node_color = QtGui.QColor(constants.DEFAULT_NODE_COLOR)
        for lib in all_classes:
            for module in all_classes[lib]:
                classes = all_classes[lib][module]["classes"]
                for c in classes:
                    c_name, _ = c
                    if c_name == self.logic_node.class_name:
                        node_color = QtGui.QColor(all_classes[lib][module]["color"])
                        node_color.setAlphaF(0.8)
                        return node_color

        return node_color

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        qp = QtGui.QPainter()
        qp.begin(self)

        # Fill
        path = QtGui.QPainterPath()
        path.addRoundedRect(QtCore.QRect(0, 0, self.width(), self.height() - 2), 12, 12)
        linearGrad = QtGui.QLinearGradient(0, 0, 0, self.height())
        linearGrad.setColorAt(0.0, QtGui.QColor(255, 255, 255, 50))
        linearGrad.setColorAt(0.1, QtGui.QColor(255, 255, 255, 0))
        linearGrad.setColorAt(0.7, QtGui.QColor(255, 255, 255, 0))
        linearGrad.setColorAt(1.0, QtGui.QColor(0, 0, 0, 50))
        qp.fillPath(path, self.color)
        qp.fillPath(path, linearGrad)
        qp.fillPath(
            path, QtGui.QBrush(QtGui.QColor(0, 0, 0, 50), QtCore.Qt.BDiagPattern)
        )

        # Shadow
        qp.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 130), 5))
        qp.drawLine(QtCore.QLine(12, self.height(), self.width() - 12, self.height()))

        # End
        qp.end()
        QtWidgets.QWidget.paintEvent(self, event)

    def clear(self):
        """
        Clear all contents of this widget.
        """
        for i in reversed(range(self.grid_layout.count())):
            w = self.grid_layout.itemAt(i).widget()
            if w:
                w.deleteLater()

    def represent_node(self):
        """
        Add the needed child widgets to properly represent all the
        internal information in the node.
        """
        self.row_count = 0

        name_label = QtWidgets.QLabel(self.logic_node.full_name)
        name_label.setFont(QtGui.QFont("arial", 14))
        self.grid_layout.addWidget(name_label, self.row_count, 0, 1, 3)
        close_btn = QtWidgets.QToolButton()
        close_btn.setIcon(QtGui.QIcon("icons:close.svg"))
        close_btn.setToolTip("Close this node panel")
        close_btn.clicked.connect(self.deleteLater)
        self.grid_layout.addWidget(close_btn, self.row_count, 2, 1, 1)
        self.row_count += 1

        # IN attributes
        in_attrs = self.logic_node.get_input_attrs()
        if in_attrs:
            in_l = QtWidgets.QLabel("Inputs")
            in_l.setAlignment(QtCore.Qt.AlignVCenter)
            self.grid_layout.addWidget(in_l, self.row_count, 0, 1, 3)
            self.row_count += 1

        for attr in in_attrs:
            self.represent_attribute(attr)

        # OUT attributes
        out_attrs = self.logic_node.get_output_attrs()
        if out_attrs:
            line = QtWidgets.QFrame()
            line.setFrameShape(QtWidgets.QFrame.HLine)
            self.grid_layout.addWidget(line, self.row_count, 0, 1, 3)
            self.row_count += 1

            out_l = QtWidgets.QLabel("Ouputs")
            out_l.setAlignment(QtCore.Qt.AlignVCenter)
            self.grid_layout.addWidget(out_l, self.row_count, 0, 1, 3)
            self.row_count += 1

        for attr in out_attrs:
            self.represent_attribute(attr)

        # LOGS
        fail_log = self.logic_node.fail_log
        error_log = self.logic_node.error_log

        if error_log or fail_log:
            line = QtWidgets.QFrame()
            line.setFrameShape(QtWidgets.QFrame.HLine)
            self.grid_layout.addWidget(line, self.row_count, 0, 1, 3)
            self.row_count += 1

        if fail_log:
            fail_label = QtWidgets.QTextEdit()
            fail_label.setReadOnly(True)
            fail_label.setText("<b>Fail log:</b><br>" + "<br>".join(fail_log))
            fail_label.setStyleSheet(
                "color:orange; background-color:rgba(0,0,0,50); border: 1px solid rgba(200,200,200,50); border-radius:5px;"
            )
            self.grid_layout.addWidget(fail_label, self.row_count, 0, 1, 3)
            self.row_count += 1

        if error_log:
            error_label = QtWidgets.QTextEdit()
            error_label.setReadOnly(True)
            error_label.setText("<b>Error log:</b><br>" + "<br>".join(error_log))
            error_label.setStyleSheet(
                "color:red; background-color:rgba(0,0,0,50); border: 1px solid rgba(200,200,200,50); border-radius:5px;"
            )
            self.grid_layout.addWidget(error_label, self.row_count, 0, 1, 3)
            self.row_count += 1

        # Execution time
        exec_time = QtWidgets.QLabel(
            f"Execution time: {self.logic_node.execution_time:.4f} s."
        )
        self.grid_layout.addWidget(exec_time, self.row_count, 0, 1, 3)
        self.row_count += 1

        # Spacer
        verticalSpacer = QtWidgets.QSpacerItem(
            20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        self.grid_layout.addItem(verticalSpacer, self.row_count, 0, 1, 3)

    def represent_attribute(self, attr):
        """
        Add one of the node's attributes to the widget.

        It will consist of different parts: a name label, an input text line...

        Args:
            attr (GeneralLogicAttribute): attr to be represented

        """
        attr_name = attr.attribute_name
        data_type = attr.data_type
        name_l = QtWidgets.QLabel("{} ({})".format(attr_name, attr.get_datatype_str()))
        name_l.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.grid_layout.addWidget(name_l, self.row_count, 0, 1, 1)

        value = attr.value
        attrib_value_le = QtWidgets.QLineEdit()
        attrib_value_le.setPlaceholderText("None")

        if value is not None:
            attrib_value_le.setText(str(value))

        regex = QtCore.QRegExp(".*")
        if data_type is str and value == "":
            attrib_value_le.setText("''")
        elif data_type is float:
            regex = QtCore.QRegExp("\d+\.\d+")
        elif data_type is int:
            regex = QtCore.QRegExp("\d+")
        elif data_type is dict:
            regex = QtCore.QRegExp("\{.+")
        elif data_type is bool:
            regex = QtCore.QRegExp("[01]")
            if value is not None:
                attrib_value_le.setText(str(int(value)))

        validator = QtGui.QRegExpValidator(regex)
        attrib_value_le.setValidator(validator)
        attrib_value_le.setToolTip("Valid input: {}".format(regex.pattern()))

        connector_type = attr.connector_type
        conns = attr.get_connections_list()
        if (
            connector_type == constants.OUTPUT
            or attr_name
            in [
                constants.START,
                constants.COMPLETED,
            ]
            or conns
        ):
            attrib_value_le.setToolTip("Does not accept manual input")
            attrib_value_le.setReadOnly(True)
            if conns:
                attrib_value_le.setToolTip("Is connected, does not accept manual input")
            attrib_value_le.setStyleSheet("border: 1px solid white")
        else:
            attrib_value_le.setStyleSheet(
                "QLineEdit{color:cyan;} QLineEdit:!focus{border: 1px solid white}\
                                            QLineEdit:focus{border: 1px dashed white}"
            )
            attrib_value_le.editingFinished.connect(
                lambda: self.logic_node.set_attribute_from_str(
                    attr_name, attrib_value_le.text()
                )
            )

        self.grid_layout.addWidget(attrib_value_le, self.row_count, 1, 1, 1)

        if attr.get_datatype_str() not in [
            "Run",
            "RunLoop",
        ]:
            copy_btn = QtWidgets.QToolButton()
            copy_btn.setIcon(QtGui.QIcon("icons:clipboard.svg"))
            copy_btn.setToolTip("Copy this value to clipboard")
            copy_btn.clicked.connect(
                lambda: QtWidgets.QApplication.clipboard().setText(
                    attrib_value_le.text()
                )
            )
            self.grid_layout.addWidget(copy_btn, self.row_count, 2, 1, 1)

        self.row_count += 1

        if conns:
            for conn in conns:
                conn_l = QtWidgets.QLabel("└─")
                conn_l.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
                self.grid_layout.addWidget(conn_l, self.row_count, 0, 1, 1)
                attrib_conns = QtWidgets.QLabel()
                if connector_type == constants.OUTPUT:
                    attrib_conns.setText(conn[1])
                else:
                    attrib_conns.setText(conn[0])
                self.grid_layout.addWidget(attrib_conns, self.row_count, 1, 1, 1)
                self.row_count += 1


class AttributeEditor(QtWidgets.QScrollArea):
    def __init__(self):
        QtWidgets.QScrollArea.__init__(self)
        self.layout = QtWidgets.QVBoxLayout()
        self.widget = QtWidgets.QWidget()
        self.widget.setLayout(self.layout)
        self.setWidgetResizable(True)
        self.setWidget(self.widget)
        self.layout.addItem(
            QtWidgets.QSpacerItem(
                20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
            ),
        )

        self.setMinimumWidth(360)

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

    def add_node_panel(self, logic_node: GeneralLogicNode):
        """
        Add a logic node to the attribute editor.

        Args:
            logic_node (GeneralLogicNode): logic node to be represented
        """
        for i in range(self.layout.count()):
            w = self.layout.itemAt(i).widget()
            if w:
                if w.logic_node == logic_node:
                    self.layout.takeAt(i)
                    self.layout.insertWidget(0, w)
                    self.verticalScrollBar().setValue(0)
                    return

        self.layout.insertWidget(0, NodePanel(logic_node))

    def refresh_node_panel(self, logic_node: GeneralLogicNode):
        """
        Refresh the representation/widget of a logic node in attribute editor.

        Args:
            logic_node (GeneralLogicNode): logic node to be refreshed
        """
        for i in range(self.layout.count()):
            w = self.layout.itemAt(i).widget()
            if w:
                if w.logic_node == logic_node:
                    w.clear()
                    w.represent_node()
                    return

    def remove_node_panel(self, logic_node: GeneralLogicNode):
        """
        Remove the representation/widget of a logic node in attribute editor.

        Args:
            logic_node (GeneralLogicNode): logic node to be removed
        """
        for i in range(self.layout.count()):
            w = self.layout.itemAt(i).widget()
            if w:
                if w.logic_node == logic_node:
                    w.deleteLater()
                    return

    def refresh(self):
        """
        Perform an overall refresh of all representations of nodes.
        """
        for i in range(self.layout.count()):
            w = self.layout.itemAt(i).widget()
            if w:
                w.clear()
                w.represent_node()

    def clear_all(self):
        """
        Remove all representations of nodes.
        """
        for i in range(self.layout.count()):
            w = self.layout.itemAt(i).widget()
            if w:
                w.deleteLater()



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import polars as pl

from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class PolarsCsv(GeneralLogicNode):
    INPUTS_DICT = {
        "in_csv_filepath": {"type": str},
    }

    OUTPUTS_DICT = {"out_dataframe": {"type": pl.DataFrame}}

    def run(self):
        import os

        in_csv_filepath = self.get_attribute_value("in_csv_filepath")

        if not os.path.isfile(in_csv_filepath):
            self.fail("CSV file {} does not exist!".format(in_csv_filepath))
            return

        self.set_output("out_dataframe", pl.read_csv(in_csv_filepath))


class PolarsParquet(GeneralLogicNode):
    INPUTS_DICT = {
        "in_parquet_filepath": {"type": str},
    }

    OUTPUTS_DICT = {"out_dataframe": {"type": pl.DataFrame}}

    def run(self):
        import os

        in_parquet_filepath = self.get_attribute_value("in_parquet_filepath")

        if not os.path.isfile(in_parquet_filepath):
            self.fail("Parquet file {} does not exist!".format(in_parquet_filepath))
            return

        self.set_output("out_dataframe", pl.read_parquet(in_parquet_filepath))



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class MirrorFolders(GeneralLogicNode):
    NICE_NAME = "Mirror folders"

    INPUTS_DICT = {
        "src_folder": {"type": str},
        "dest_folder": {"type": str},
    }

    def run(self):
        import platform
        import subprocess

        src_folder = self.get_attribute_value("src_folder")
        dest_folder = self.get_attribute_value("dest_folder")

        if platform.system() == "Windows":
            s = subprocess.Popen(["robocopy", src_folder, dest_folder, "/E", "/MIR"])
            s.wait()
        else:
            self.fail("Not implemented!")


class GetFolderFromFilepath(GeneralLogicNode):
    INPUTS_DICT = {
        "filepath": {"type": str},
    }

    OUTPUTS_DICT = {"folder_path": {"type": str}}

    def run(self):
        import os

        filepath = self.get_attribute_value("filepath")
        if not filepath:
            self.fail("No filepath provided!")
            return

        self.set_output("folder_path", os.path.dirname(filepath))


class PathJoin(GeneralLogicNode):
    INPUTS_DICT = {
        "token_0": {"type": str},
        "token_1": {"type": str},
    }

    OUTPUTS_DICT = {"joined_path": {"type": str}}

    def run(self):
        import os

        token_0 = self.get_attribute_value("token_0")
        token_1 = self.get_attribute_value("token_1")
        self.set_output("joined_path", os.path.join(token_0, token_1))


class CopyFile(GeneralLogicNode):
    INPUTS_DICT = {
        "source_file": {"type": str},
        "destination_file": {"type": str},
        "make_dirs": {"type": bool, "optional": True},
    }

    def run(self):
        import os
        import shutil

        source_file = self.get_attribute_value("source_file")
        destination_file = self.get_attribute_value("destination_file")
        make_dirs = self.get_attribute_value("make_dirs")
        if make_dirs:
            os.makedirs(os.path.dirname(destination_file), exist_ok=True)
        shutil.copy(source_file, destination_file)


class CopyFoldersRecursive(GeneralLogicNode):
    NICE_NAME = "Copy folders recursive"

    INPUTS_DICT = {
        "source_folder_path": {"type": str},
        "destination_folder_path": {"type": str},
    }



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class JsonToDict(GeneralLogicNode):
    NICE_NAME = "Json to dict"
    HELP = "Load a JSON file"

    INPUTS_DICT = {
        "in_json_filepath": {"type": str},
    }

    OUTPUTS_DICT = {"out_dict": {"type": dict}}

    def run(self):
        import os
        import json

        in_json_filepath = self.get_attribute_value("in_json_filepath")

        if not os.path.isfile(in_json_filepath):
            self.fail("Json file {} does not exist!".format(in_json_filepath))
            return

        with open(in_json_filepath, "r") as json_file:
            data = json.load(json_file)
            self.set_output("out_dict", data)


class TxtToList(GeneralLogicNode):
    INPUTS_DICT = {
        "in_txt": {"type": str},
    }
    OUTPUTS_DICT = {"out_list": {"type": list}}

    def run(self):
        in_txt = self.get_attribute_value("in_txt")
        with open(in_txt, "r") as file:
            self.set_output("out_list", file.readlines())


class YamlToDict(GeneralLogicNode):
    INPUTS_DICT = {
        "yaml_filepath": {"type": str},
    }

    OUTPUTS_DICT = {"out_dict": {"type": dict}}

    def run(self):
        import os
        import yaml

        yaml_filepath = self.get_attribute_value("yaml_filepath")

        if not os.path.isfile(yaml_filepath):
            self.fail("Yaml file {} does not exist!".format(yaml_filepath))
            return

        with open(yaml_filepath, "r") as y:
            self.set_output("out_dict", yaml.safe_load(y))


class CsvToTable(GeneralLogicNode):
    INPUTS_DICT = {
        "csv_filepath": {"type": str},
        "delimiter": {"type": str, "optional": True},
    }

    OUTPUTS_DICT = {"out_table": {"type": list}}

    def run(self):
        import os
        import csv

        csv_filepath = self.get_attribute_value("csv_filepath")
        delimiter = self.get_attribute_value("delimiter") or ","

        if not os.path.isfile(csv_filepath):
            self.fail("CSV file {} does not exist!".format(csv_filepath))
            return

        out_table = []
        with open(csv_filepath, newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=delimiter, quotechar="|")
            for row in reader:
                out_table.append(row)

        self.set_output("out_table", out_table)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtSvg
from PySide2 import QtWidgets

from all_nodes import constants
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


ANNOTATION_TYPES_WITH_TEXT = [
    "note",
    "paper",
    "long_tag",
    "short_tag",
    "notepad",
    "post_it",
]


# -------------------------------- ANNOTATION CLASS -------------------------------- #
class GeneralGraphicAnnotation(QtWidgets.QGraphicsPathItem):
    def __init__(self, annotation_type: str):
        # INIT
        QtWidgets.QGraphicsPathItem.__init__(self)
        self.setData(0, constants.GRAPHIC_ANNOTATION)
        self.setAcceptHoverEvents(True)
        self.setFlags(
            QtWidgets.QGraphicsPathItem.ItemIsMovable
            | QtWidgets.QGraphicsPathItem.ItemIsSelectable
            | QtWidgets.QGraphicsPathItem.ItemSendsScenePositionChanges
        )

        self.setPen(QtCore.Qt.NoPen)

        # Setup graphics
        self.annotation_type = annotation_type
        self.width = 0
        self.height = 0
        self.setup_graphics()

    def setup_graphics(self):
        # Main graphics
        self.renderer = QtSvg.QSvgRenderer(f"graphics:{self.annotation_type}.svg")
        main_graphics = QtSvg.QGraphicsSvgItem(parentItem=self)
        main_graphics.setSharedRenderer(self.renderer)

        self.width = main_graphics.boundingRect().width()
        self.height = main_graphics.boundingRect().height()

        padding = 40

        # Text input
        self.proxy_edit = QtWidgets.QGraphicsProxyWidget(parent=self)
        self.note_text_edit = QtWidgets.QPlainTextEdit()
        self.note_text_edit.setFont(QtGui.QFont("arial", 8))
        self.note_text_edit.setStyleSheet(
            "color:black; background-color:transparent; border:1px dotted rgba(100,100,100,120);"
        )
        self.note_text_edit.setFixedSize(self.width - padding, self.height - padding)
        self.note_text_edit.setPlainText("Write here...")
        self.proxy_edit.setWidget(self.note_text_edit)
        self.proxy_edit.setPos(
            self.width,
            self.height,
        )

        if self.annotation_type not in ANNOTATION_TYPES_WITH_TEXT:
            self.proxy_edit.hide()

        # Main shape
        path = QtGui.QPainterPath()
        path.addRect(
            QtCore.QRect(
                0,
                0,
                self.width + padding,
                self.height + padding,
            ),
        )
        main_graphics.setPos(padding / 2, padding / 2)
        self.proxy_edit.setPos(padding, padding)
        self.setPath(path)

    def get_type(self):
        return self.annotation_type

    def get_text(self):
        if self.annotation_type in ANNOTATION_TYPES_WITH_TEXT:
            return self.note_text_edit.toPlainText()
        return None

    def set_text(self, text: str):
        self.note_text_edit.setPlainText(text)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import unittest

from all_nodes import constants
from all_nodes.logic.logic_scene import LogicScene
from all_nodes.logic.logic_scene import LogicSceneError
from all_nodes import utils


# -------------------------------- TESTS -------------------------------- #
class ContextTesting(unittest.TestCase):
    def test_context_creation(self):
        """
        Create a context.
        """
        utils.print_test_header("test_context_creation")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("EnvironToYmlCtx")
        self.assertIsNotNone(n_1)

    def test_fake_context_creation(self):
        """
        Try to create a context that does not exist.
        """
        utils.print_test_header("test_fake_context_creation")

        logic_scene = LogicScene()
        with self.assertRaises(LogicSceneError) as e:
            logic_scene.add_node_by_name("FakeCtx")
        print(e.exception)

    def test_context_run(self):
        """
        Create and run a context.
        """
        utils.print_test_header("test_context_run")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("EnvironToYmlCtx")
        n_1.run_single()
        self.assertEqual(n_1.success, constants.SUCCESSFUL)

    def test_run_chain_from_ctx(self):
        """
        Run a chain of nodes starting at a context.
        """
        utils.print_test_header("test_run_chain_from_ctx")

        logic_scene = LogicScene()

        n_1 = logic_scene.add_node_by_name("EnvironToYmlCtx")
        n_2 = logic_scene.add_node_by_name("EmptyNode")
        n_1["COMPLETED"].connect_to_other(n_2["START"])

        n_1.run_chain()



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.constants import PreviewsGUI
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class StringPreview(GeneralLogicNode):
    INPUTS_DICT = {
        "in_str": {"type": str},
    }

    INTERNALS_DICT = {
        "internal_str": {"type": str, "gui_type": PreviewsGUI.STR_PREVIEW},
    }

    def run(self):
        self.set_attribute_value("internal_str", self.get_attribute_value("in_str"))


class MultilineStringPreview(GeneralLogicNode):
    INPUTS_DICT = {
        "in_str": {"type": str},
    }

    INTERNALS_DICT = {
        "internal_multiline_str": {
            "type": str,
            "gui_type": PreviewsGUI.MULTILINE_STR_PREVIEW,
        },
    }

    def run(self):
        self.set_attribute_value(
            "internal_multiline_str", self.get_attribute_value("in_str")
        )


class DictPreview(GeneralLogicNode):
    INPUTS_DICT = {
        "in_dict": {"type": dict},
    }

    INTERNALS_DICT = {
        "internal_multiline_dict": {
            "type": dict,
            "gui_type": PreviewsGUI.DICT_PREVIEW,
        },
    }

    def run(self):
        self.set_attribute_value(
            "internal_multiline_dict", self.get_attribute_value("in_dict")
        )



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import concurrent.futures
import importlib
import inspect
import os
from pathlib import Path
import time

from PySide2 import QtCore
import yaml

from all_nodes import constants
from all_nodes import utils
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes.graphic.widgets.global_signaler import GlobalSignaler


GS = GlobalSignaler()

LOGGER = utils.get_logger(__name__)


# -------------------------------- NODE CLASSES -------------------------------- #
CLASSES_TO_SKIP = [
    "InputsGUI",
    "PreviewsGUI",
    "GeneralLogicNode",
    "Run",
    "RunLoop",
    "GrabInputFromCtx",
    "SetOutputToCtx",
]  # Classes to skip when gathering all usable clases, populating widgets...


def register_node_lib(lib_path):
    classes_dict = dict()

    all_py = []
    for root, _, files in os.walk(lib_path, topdown=True):
        for file in files:
            p = os.path.join(root, file)
            if p.endswith(".py") and "__init__" not in p:
                all_py.append(p)

    for py_path in all_py:
        node_library_path = os.path.dirname(py_path)
        node_library_name = os.path.basename(node_library_path)
        module_filename = os.path.basename(py_path)
        module_name = os.path.splitext(module_filename)[0]
        icons_path = os.path.join(node_library_path, "icons")
        styles_path = os.path.join(node_library_path, "styles.yml")

        if classes_dict.get(node_library_name) is None:
            classes_dict[node_library_name] = dict()

        # ICONS - registering the icons so they can be found
        if not os.path.isdir(icons_path):
            LOGGER.warning(
                f"No icons folder available for {node_library_name}, icons for this library should be saved at: {icons_path}"
            )
        if os.path.isdir(icons_path) and icons_path not in QtCore.QDir.searchPaths(
            "icons"
        ):
            QtCore.QDir.addSearchPath("icons", icons_path)
            LOGGER.debug("Registered path {} to 'icons'".format(icons_path))

        # STYLES
        node_styles = dict()
        if not os.path.isfile(styles_path):
            LOGGER.warning(
                f"No styles file available for {node_library_name}, styles for this library should be saved at: {styles_path}"
            )
        else:
            with open(styles_path, "r") as stream:
                node_styles = yaml.safe_load(stream)

        # CLASSES SCANNING
        loaded_spec = importlib.util.spec_from_file_location(
            module_name,
            py_path,
        )
        loaded_module = importlib.util.module_from_spec(loaded_spec)
        loaded_spec.loader.exec_module(loaded_module)
        class_members = inspect.getmembers(loaded_module, inspect.isclass)
        if not class_members:
            return

        classes_dict[node_library_name][module_name] = dict()
        module_classes = list()
        class_counter = 0
        for name, cls_object in class_members:
            if (
                not issubclass(cls_object, GeneralLogicNode)
                or cls_object == GeneralLogicNode
            ):
                continue

            # Icon for this class  # TODO Refactor this out
            default_icon = node_styles.get(module_name, dict()).get("default_icon")
            icon_path = "icons:nodes.svg"
            if (
                hasattr(cls_object, "IS_CONTEXT") and cls_object.IS_CONTEXT
            ):  # TODO inheritance not working here?
                icon_path = "icons:cubes.svg"
            if QtCore.QFile.exists(f"icons:{name}.png"):
                icon_path = f"icons:{name}.png"
            elif QtCore.QFile.exists(f"icons:{name}.svg"):
                icon_path = f"icons:{name}.svg"
            elif default_icon:
                if QtCore.QFile.exists("icons:" + default_icon + ".png"):
                    icon_path = f"icons:{default_icon}.png"
                elif QtCore.QFile.exists("icons:" + default_icon + ".svg"):
                    icon_path = f"icons:{default_icon}.svg"
            setattr(cls_object, "ICON_PATH", icon_path)

            # Class name and object
            setattr(cls_object, "FILEPATH", py_path)  # TODO not ideal?
            module_classes.append((name, cls_object))
            class_counter += 1

        classes_dict[node_library_name][module_name]["node_lib_path"] = (
            node_library_path
        )
        classes_dict[node_library_name][module_name]["node_lib_name"] = (
            node_library_name
        )
        classes_dict[node_library_name][module_name]["module_filename"] = (
            module_filename
        )
        classes_dict[node_library_name][module_name]["module_full_path"] = py_path
        classes_dict[node_library_name][module_name]["classes"] = module_classes

        classes_dict[node_library_name][module_name]["color"] = (
            constants.DEFAULT_NODE_COLOR
        )
        for module_style in node_styles:
            if module_style in module_name:
                classes_dict[node_library_name][module_name]["color"] = node_styles[
                    module_style
                ].get("color", constants.DEFAULT_NODE_COLOR)
        LOGGER.debug(
            "Scanned {} for classes: found {}".format(
                os.path.basename(py_path), class_counter
            )
        )

    return classes_dict


def get_all_node_libs():
    # TODO clarify this naming better (maybe project->lib->module?)
    # Paths to be examined
    libraries_path = [
        lib_path
        for lib_path in os.getenv("ALL_NODES_LIB_PATH", "").split(os.pathsep)
        if lib_path
    ]
    if not os.getenv("ALL_NODES_LIB_PATH"):
        root = os.path.abspath(__file__)
        root_dir_path = os.path.dirname(os.path.dirname(root))
        default_lib_path = os.path.join(root_dir_path, "lib", "base_node_lib")
        libraries_path = [default_lib_path]
        LOGGER.warning(
            "Env variable 'ALL_NODES_LIB_PATH' is not defined, "
            "will just scan for node libraries at default location: " + default_lib_path
        )

    return libraries_path


# -------------------------------- SCENES -------------------------------- #
def get_all_scenes_recursive(libraries_path=None, scenes_dict=None):
    if libraries_path is None:
        libraries_path = []
        if not os.getenv("ALL_NODES_LIB_PATH"):
            root = os.path.abspath(__file__)
            root_dir_path = os.path.dirname(os.path.dirname(root))
            default_lib_path = os.path.join(
                root_dir_path, "lib", "basic_examples_scene_lib"
            )
            libraries_path.append(default_lib_path)
            LOGGER.warning(
                "Env variable 'ALL_NODES_LIB_PATH' is not defined, "
                "will just scan for scene libraries at default location: "
                + default_lib_path
            )
        else:
            all_paths = os.getenv("ALL_NODES_LIB_PATH").split(os.pathsep)
            for path in all_paths:
                path = path.strip()
                if path:
                    for elem in os.listdir(path):
                        full_path = os.path.join(path, elem)
                        if os.path.isdir(full_path) and "scene_lib" in full_path:
                            libraries_path.append(full_path)

    if scenes_dict is None:
        scenes_dict = dict()

    for path in libraries_path:
        folder_name = os.path.basename(os.path.normpath(path))
        if not os.path.exists(path):
            LOGGER.error(
                "Path {} registered to ALL_NODES_LIB_PATH does not exist!".format(path)
            )
        if not os.listdir(path):
            continue
        scenes_dict[folder_name] = list()
        for elem in os.listdir(path):
            full_path = os.path.join(path, elem)
            scene_name = os.path.splitext(elem)[0]
            if os.path.isfile(full_path) and full_path.endswith(".yml"):
                scenes_dict[folder_name].append((scene_name, full_path))
            elif os.path.isdir(full_path):
                new_dict = dict()
                scenes_dict[folder_name].append(new_dict)
                get_all_scenes_recursive([full_path], new_dict)

    return scenes_dict


def get_scene_from_alias(scenes_dict, alias):
    """
    Given just an alias (or more specifically filename) find full path to the yml file.
    Scans recursively.

    Args:
        scenes_dict (dict): dict with all the scenes and their paths
        alias (str): alias/filename to search for

    Returns: str, full path to the yml scene found
    """
    scene_path = None

    for key in scenes_dict:
        list_value = scenes_dict[key]
        for elem in list_value:
            if isinstance(elem, dict):
                scene_path = get_scene_from_alias(elem, alias)
                if scene_path is not None:
                    return scene_path
            else:
                for elem in list_value:
                    if isinstance(elem, dict):
                        scene_path = get_scene_from_alias(elem, alias)
                        if scene_path is not None:
                            return scene_path
                    elif isinstance(elem, tuple):
                        scene_name, full_path = elem
                        if scene_name == alias:
                            scene_path = full_path
                            return scene_path
    return scene_path


# -------------------------------- Class Registry -------------------------------- #
class ClassRegistry:
    # Instance
    _instance = None

    # Attributes
    _all_classes = None
    _all_scenes = None

    _all_classes_simplified = None

    # Workers
    _lib_workers = []
    _time_start = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ClassRegistry, cls).__new__(cls)
        return cls._instance

    def get_workers(cls):
        return cls._lib_workers

    def scan_for_classes_GUI(cls):
        """
        Scan for classes in GUI mode.

        In this approach, all the node libraries are scanned in parallel by using worker threads. The signals of these
        will tell the UI when to populate its elements, so the main UI thread is not blocked while scanning.
        """
        LOGGER.info("Gathering all classes (GUI mode)...")
        cls._time_start = time.time()
        cls._all_classes = dict()

        node_libs = get_all_node_libs()
        cls._lib_workers = [LibWorker(node_lib) for node_lib in node_libs]
        for worker in cls._lib_workers:
            QtCore.QThreadPool.globalInstance().start(worker)
            worker.signaler.finished.connect(cls.update_classes_dict)

    def scan_for_classes(cls):
        LOGGER.info("Gathering all classes...")
        t1 = time.time()  # TODO find something more precise
        cls._all_classes = dict()

        with concurrent.futures.ThreadPoolExecutor(10) as executor:
            futures = [
                executor.submit(register_node_lib, full_path)
                for full_path in get_all_node_libs()
            ]
            for future in concurrent.futures.as_completed(futures):
                cls._all_classes.update(future.result())

        LOGGER.info(f"Total time scanning classes: {time.time() - t1}s.")
        GS.signals.class_scanning_finished.emit()

    def get_all_classes(cls):
        if cls._all_classes is None:
            cls.scan_for_classes()
        return cls._all_classes

    def get_all_scenes(cls):
        if cls._all_scenes is None:
            cls._all_scenes = get_all_scenes_recursive()

        return cls._all_scenes

    def get_all_classes_simplified(cls):
        """
        Get a simplified list of all classes and their corresponding icon paths.

        Returns:
            list: A list of tuples containing the class name and its icon path.
        """
        if cls._all_classes_simplified is None:
            cls._all_classes_simplified = list()
            all_classes = cls.get_all_classes()
            for lib in sorted(all_classes):
                for m in all_classes[lib]:
                    for name, class_object in all_classes[lib][m]["classes"]:
                        cls._all_classes_simplified.append(
                            (name, class_object.ICON_PATH)
                        )

        return cls._all_classes_simplified

    def get_icon_path(cls, class_name_to_search: str):
        """
        Get the icon path for a given class name.

        Args:
            class_name_to_search (str): The name of the class to search for.

        Returns:
            str: The icon path of the class if found, None otherwise.
        """
        for class_name, icon_path in cls.get_all_classes_simplified():
            if class_name_to_search == class_name:
                return icon_path

    def update_classes_dict(cls):
        for worker in cls._lib_workers:
            if worker.finished:
                cls._all_classes.update(worker.dict_lib)
                cls._lib_workers.remove(worker)

        if not len(cls._lib_workers):
            LOGGER.info(
                f"Total time scanning classes: {time.time() - cls._time_start}s."
            )
            GS.signals.class_scanning_finished.emit()

    def flush(cls):
        """
        Resets all class-related attributes to None.
        """
        cls._all_classes = None
        cls._all_scenes = None
        cls._all_classes_simplified = None

        cls._lib_workers = []


CLASS_REGISTRY = ClassRegistry()  # Singleton to use


# -------------------------------- WORKER -------------------------------- #
class LibWorkerSignaler(QtCore.QObject):
    finished = QtCore.Signal()


class LibWorker(QtCore.QRunnable):
    def __init__(self, lib_path: str):
        super(LibWorker, self).__init__()

        self.lib_path = lib_path
        self.lib_name = Path(lib_path).name
        self.dict_lib = None

        self.signaler = LibWorkerSignaler()

        self.finished = False

    def run(self):
        path = self.lib_path.strip()

        # TODO examine classes to make sure there are no repeated names?
        self.dict_lib = register_node_lib(path)
        self.finished = True
        self.signaler.finished.emit()



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class BooleanOr(GeneralLogicNode):
    NICE_NAME = "Or"
    HELP = "Return True if at least one input is True"

    INPUTS_DICT = {
        "in_0": {"type": bool},
        "in_1": {"type": bool},
    }

    OUTPUTS_DICT = {"out": {"type": bool}}

    def run(self):
        in_0 = self.get_attribute_value("in_0")
        in_1 = self.get_attribute_value("in_1")
        self.set_output("out", in_0 or in_1)


class BooleanAnd(GeneralLogicNode):
    NICE_NAME = "And"
    HELP = "Return True if all inputs are True"

    INPUTS_DICT = {
        "in_0": {"type": bool},
        "in_1": {"type": bool},
    }

    OUTPUTS_DICT = {"out": {"type": bool}}

    def run(self):
        in_0 = self.get_attribute_value("in_0")
        in_1 = self.get_attribute_value("in_1")
        self.set_output("out", in_0 and in_1)


class BooleanXor(GeneralLogicNode):
    NICE_NAME = "Xor"
    HELP = "Return True if only one input is True"

    INPUTS_DICT = {
        "in_0": {"type": bool},
        "in_1": {"type": bool},
    }

    OUTPUTS_DICT = {"out": {"type": bool}}

    def run(self):
        in_0 = self.get_attribute_value("in_0")
        in_1 = self.get_attribute_value("in_1")
        self.set_output("out", in_0 != in_1)


class BooleanNot(GeneralLogicNode):
    NICE_NAME = "Not"
    HELP = "Return the opposite of the input"

    INPUTS_DICT = {
        "in_0": {"type": bool},
    }

    OUTPUTS_DICT = {"out": {"type": bool}}

    def run(self):
        in_0 = self.get_attribute_value("in_0")
        self.set_output("out", not in_0)


class BooleanNand(GeneralLogicNode):
    NICE_NAME = "Nand"
    HELP = "Return True if not all inputs are True"

    INPUTS_DICT = {
        "in_0": {"type": bool},
        "in_1": {"type": bool},
    }

    OUTPUTS_DICT = {"out": {"type": bool}}

    def run(self):
        in_0 = self.get_attribute_value("in_0")
        in_1 = self.get_attribute_value("in_1")
        self.set_output("out", not (in_0 and in_1))


class BooleanNXor(GeneralLogicNode):
    NICE_NAME = "NXor"
    HELP = "Return True if both inputs are the same"

    INPUTS_DICT = {
        "in_0": {"type": bool},
        "in_1": {"type": bool},
    }

    OUTPUTS_DICT = {"out": {"type": bool}}

    def run(self):
        in_0 = self.get_attribute_value("in_0")
        in_1 = self.get_attribute_value("in_1")
        self.set_output("out", in_0 == in_1)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class EnvironToYmlCtx(GeneralLogicNode):
    IS_CONTEXT = True
    OUTPUTS_DICT = {"yaml_filepath": {"type": str}}



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class StringSplit(GeneralLogicNode):
    NICE_NAME = "String split"

    INPUTS_DICT = {
        "in_str": {"type": str},
        "split_char": {"type": str, "optional": True},
    }

    OUTPUTS_DICT = {"out_list": {"type": list}}

    def run(self):
        in_str = self.get_attribute_value("in_str")
        split_char = self.get_attribute_value("split_char") or " "

        split_list = in_str.split(split_char)
        if split_char == "\\n":
            split_list = in_str.splitlines()

        self.set_output("out_list", split_list)


class StringFormat(GeneralLogicNode):
    INPUTS_DICT = {
        "in_str": {"type": str},
        "format_dict": {"type": dict},
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        in_str = self.get_attribute_value("in_str")
        format_dict = self.get_attribute_value("format_dict")

        self.set_output("out_str", in_str.format(**format_dict))



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import PIL

from all_nodes.constants import InputsGUI, PreviewsGUI
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class PIL_ImageOpen(GeneralLogicNode):
    INPUTS_DICT = {"in_path": {"type": str}}
    OUTPUTS_DICT = {"out_image": {"type": PIL.Image.Image}}

    def run(self):
        self.set_output(
            "out_image",
            PIL.Image.open(self.get_attribute_value("in_path")),
        )


class PIL_ImagePreview(GeneralLogicNode):
    INPUTS_DICT = {"in_image": {"type": PIL.Image.Image}}

    INTERNALS_DICT = {
        "internal_image": {
            "type": PIL.Image.Image,
            "gui_type": PreviewsGUI.IMAGE_PREVIEW,
        }
    }

    def run(self):
        self.set_attribute_value("internal_image", self.get_attribute_value("in_image"))


class PIL_DirectOpen(GeneralLogicNode):
    OUTPUTS_DICT = {
        "out_image": {"type": PIL.Image.Image},
        "out_image_path": {"type": str},
    }

    INTERNALS_DICT = {
        "internal_str_image_path": {
            "type": str,
            "gui_type": InputsGUI.STR_INPUT,
        },
        "internal_image": {
            "type": PIL.Image.Image,
            "gui_type": PreviewsGUI.IMAGE_PREVIEW,
        },
    }

    def run(self):
        # Get inputs
        path = self.get_attribute_value("internal_str_image_path")
        opened_image = PIL.Image.open(path)

        # Set outputs
        self.set_output(
            "out_image",
            opened_image,
        )
        self.set_output(
            "out_image_path",
            path,
        )

        # Display previews
        self.set_attribute_value("internal_image", opened_image)


class PIL_OpenFromUrl(GeneralLogicNode):
    INPUTS_DICT = {"in_url": {"type": str}}

    OUTPUTS_DICT = {"out_image": {"type": PIL.Image.Image}}

    def run(self):
        import requests

        data = requests.get(self.get_attribute_value("in_url"), stream=True).raw
        img = PIL.Image.open(data)
        self.set_output("out_image", img)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes import constants
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes.logic.logic_node import Run, RunLoop
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class BasicIf(GeneralLogicNode):
    NICE_NAME = "Basic IF conditional"

    INPUTS_DICT = {
        "in_bool": {"type": bool},
    }

    OUTPUTS_DICT = {
        "path_1": {"type": Run, "optional": True},
        "path_2": {"type": Run, "optional": True},
    }

    def run(self):
        in_bool = self.get_attribute_value("in_bool")

        if in_bool:
            self.set_output("path_1", Run())
        else:
            self.set_output("path_2", Run())


class Basicbreaker(GeneralLogicNode):
    NICE_NAME = "Basic breaker"
    HELP = "If the input condition is met, break the whole execution after this node"

    INPUTS_DICT = {
        "in_bool": {"type": bool},
    }

    def run(self):
        in_bool = self.get_attribute_value("in_bool")

        if in_bool:
            self.error("BREAKING")


class ForEachBegin(GeneralLogicNode):
    NICE_NAME = "For each begin"

    INPUTS_DICT = {
        "iterable": {"type": list},
    }

    OUTPUTS_DICT = {
        "element": {"type": object},
        "foreach_end": {"type": RunLoop},
    }

    def _run(self, execute_connected=True):
        if not self.active:
            self.fail("Cannot start loop from an inactive loop node!")
            self.signaler.finished.emit()
            return

        from colorama import Fore, Style

        LOGGER.info(f"{Fore.CYAN}[{self.node_name}] Loop begins!{Style.RESET_ALL}")
        self.execution_counter += 1

        iterable = self.get_attribute_value("iterable")
        num_iterations = len(iterable)

        self.success = constants.IN_LOOP

        for i in range(num_iterations):
            LOGGER.info(f"{Fore.CYAN}{self.node_name}, iteration {i}{Style.RESET_ALL}")

            if execute_connected:
                for node in self.out_connected_nodes():
                    node.recursive_clear_connected_input_attrs()
                    node.recursive_set_in_loop()

            element = iterable[i]
            self.set_attribute_value("element", element)
            self.set_output(constants.COMPLETED, Run())

            self.propagate_results()
            if execute_connected:
                for node in self.out_connected_nodes():
                    if node.success in [constants.NOT_RUN, constants.IN_LOOP]:
                        LOGGER.info(
                            "[LOOP] From {}, launching execution of {}".format(
                                self.full_name, node.full_name
                            )
                        )
                        node._run()

        # Launch execution of ForEachEnd
        self.success = constants.SUCCESSFUL
        self.set_output("foreach_end", RunLoop())
        self.propagate_results()
        if execute_connected:
            for node in self.out_connected_nodes():
                if node.success == constants.IN_LOOP:
                    LOGGER.info(
                        "[LOOP END] From {}, launching execution of {}".format(
                            self.full_name, node.full_name
                        )
                    )
                    node._run()

        self.signaler.finished.emit()


class ForEachEnd(GeneralLogicNode):
    NICE_NAME = "For each end"

    INPUTS_DICT = {
        "foreach_end": {"type": RunLoop},
    }

    def run(self):
        from colorama import Fore, Style

        LOGGER.info(f"{Fore.CYAN}[{self.node_name}] Loop ended!{Style.RESET_ALL}")



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


# Nodes to grab input values from context ------------------------------------------
class GrabInputFromCtx(GeneralLogicNode):
    NICE_NAME = "CTX Input"
    HELP = "Get the value from an attribute that belongs to the parent node"

    INPUTS_DICT = {"in_parent_attr_name": {"type": str}}
    OUTPUTS_DICT = {"in_parent_attr_value": {"type": object}}

    def run(self):
        parent_attr_name = self.get_attribute_value("in_parent_attr_name")

        # Analize parent node
        parent_node = self.context
        if parent_node is None:
            self.error("This node has to be inside a context node!")
            return
        parent_attr = parent_node[parent_attr_name]
        parent_attr_value = parent_node.get_attribute_value(parent_attr_name)
        if not parent_attr:
            self.fail("Parent node has no attribute {}".format(parent_attr_name))
            return

        # Analize compatibility
        out_attr = self["in_parent_attr_value"]
        if parent_attr.data_type != out_attr.data_type:
            self.fail(
                "Parent attribute {0} ({1}) is different to output of this node {2} ({3}) : {1} != {3}".format(
                    parent_attr.dot_name,
                    parent_attr.get_datatype_str(),
                    out_attr.dot_name,
                    out_attr.get_datatype_str(),
                )
            )
            return

        # Analize connections
        for connected_attr in out_attr.connected_attributes:
            if parent_attr.data_type != connected_attr.data_type:
                self.fail(
                    "Parent attribute {0} ({1}) cannot be propagated to internal {2} ({3}) : {1} != {3}".format(
                        parent_attr.dot_name,
                        parent_attr.get_datatype_str(),
                        connected_attr.dot_name,
                        connected_attr.get_datatype_str(),
                    )
                )
                return

        self.set_output("in_parent_attr_value", parent_attr_value)


class GrabStrInputFromCtx(GrabInputFromCtx):
    NICE_NAME = "CTX Input (Str)"

    INPUTS_DICT = {"in_parent_attr_name": {"type": str}}
    OUTPUTS_DICT = {"in_parent_attr_value": {"type": str}}


class GrabDictInputFromCtx(GrabInputFromCtx):
    NICE_NAME = "CTX Input (Dict)"

    INPUTS_DICT = {"in_parent_attr_name": {"type": str}}
    OUTPUTS_DICT = {"in_parent_attr_value": {"type": dict}}


class GrabIntInputFromCtx(GrabInputFromCtx):
    NICE_NAME = "CTX Input (Int)"

    INPUTS_DICT = {"in_parent_attr_name": {"type": str}}
    OUTPUTS_DICT = {"in_parent_attr_value": {"type": int}}


class GrabBoolInputFromCtx(GrabInputFromCtx):
    NICE_NAME = "CTX Input (Bool)"

    INPUTS_DICT = {"in_parent_attr_name": {"type": str}}
    OUTPUTS_DICT = {"in_parent_attr_value": {"type": bool}}


# Nodes to set output values to context ------------------------------------------
class SetOutputToCtx(GeneralLogicNode):
    NICE_NAME = "CTX Output"

    INPUTS_DICT = {
        "out_parent_attr_name": {"type": str},
        "out_parent_attr_value": {"type": object},
    }

    def run(self):
        parent_attr_name = self.get_attribute_value("out_parent_attr_name")

        parent_node = self.context
        if parent_node is None:
            self.error("This node has to be inside a context node!")
            return
        parent_attr = parent_node[parent_attr_name]
        if not parent_attr:
            self.fail("Parent node has no attribute {}".format(parent_attr_name))
            return

        out_attr = self["out_parent_attr_value"]
        out_attr_value = self.get_attribute_value("out_parent_attr_value")
        if out_attr.data_type is not parent_attr.data_type:
            self.fail(
                "Value of {0} ({1}) cannot be propagated to parent {2} ({3}): {1} != {3}".format(
                    out_attr.dot_name,
                    out_attr.get_datatype_str(),
                    parent_attr.dot_name,
                    parent_attr.get_datatype_str(),
                )
            )
            return

        parent_node.set_output(parent_attr_name, out_attr_value)


class SetStrOutputToCtx(SetOutputToCtx):
    NICE_NAME = "CTX Output (Str)"

    INPUTS_DICT = {
        "out_parent_attr_name": {"type": str},
        "out_parent_attr_value": {"type": str},
    }


class SetDictOutputToCtx(SetOutputToCtx):
    NICE_NAME = "CTX Output (Dict)"

    INPUTS_DICT = {
        "out_parent_attr_name": {"type": str},
        "out_parent_attr_value": {"type": dict},
    }


class SetIntOutputToCtx(SetOutputToCtx):
    NICE_NAME = "CTX Output (Int)"

    INPUTS_DICT = {
        "out_parent_attr_name": {"type": str},
        "out_parent_attr_value": {"type": int},
    }


class SetBoolOutputToCtx(SetOutputToCtx):
    NICE_NAME = "CTX Output (Bool)"

    INPUTS_DICT = {
        "out_parent_attr_name": {"type": str},
        "out_parent_attr_value": {"type": bool},
    }



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import unittest

import pymongo

from all_nodes.analytics import analytics
from all_nodes.logic.logic_scene import LogicScene
from all_nodes import utils


# -------------------------------- TESTS -------------------------------- #
class StatisticsTesting(unittest.TestCase):
    def test_small_query(self):
        """
        Just make a small query to DB
        """
        utils.print_test_header("test_small_query")

        res = list(analytics.make_query({}).limit(10))

        self.assertGreater(len(res), 5)

    def test_make_wrong_query(self):
        """
        Make an impossible query DB
        """
        with self.assertRaises(pymongo.errors.OperationFailure) as e:
            analytics.make_query_aggregation(
                [
                    {
                        "$match": {
                            "success": {"$not": "FAILED"},
                            "class_name": {"$not": "666"},
                        }
                    },
                    {"$limit": 10},
                ]
            )
        print(e.exception)

    def test_submit_statistics(self):
        """
        Make a small statistics sumbission
        """
        utils.print_test_header("test_submit_statistics")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("EmptyNode")
        n_1.run_single()

        analytics.submit_bulk_analytics([n_1.get_node_full_dict()])



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.constants import InputsGUI
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class StartFile(GeneralLogicNode):
    INPUTS_DICT = {
        "file_path": {"type": str},
    }

    def run(self):
        import os
        import platform
        import subprocess

        file_path = self.get_attribute_value("file_path")
        if not os.path.exists(file_path):
            self.fail("File {} does not exist!".format(file_path))
            return

        platform_name = platform.system().lower()
        if "windows" in platform_name:
            os.startfile(file_path)
        elif platform_name in ["linux", "ubuntu"]:
            subprocess.call(["xdg-open", file_path])
        elif platform_name in ["dawrin"]:
            subprocess.call(["open", file_path])
        else:
            self.fail("Not sure how to open a file in {}...".format(platform_name))
            return


class LaunchSubprocess(GeneralLogicNode):
    INPUTS_DICT = {
        "subprocess_str": {"type": str},
    }

    def run(self):
        import subprocess

        subprocess_str = self.get_attribute_value("subprocess_str")
        subprocess.run(subprocess_str, shell=True)


class RunPython(GeneralLogicNode):
    INTERNALS_DICT = {
        "internal_python_str": {
            "type": str,
            "gui_type": InputsGUI.MULTILINE_STR_INPUT,
        },
    }

    def run(self):
        python_str = self.get_attribute_value("internal_python_str")
        exec(python_str)



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import datetime
import logging
import math
import os
import platform
import pprint
import re
import subprocess
import sys

from PySide2 import QtWidgets
from PySide2 import QtCore
from PySide2 import QtGui
import yaml

from all_nodes import constants
from all_nodes.graphic.graphic_annotation import GeneralGraphicAnnotation
from all_nodes.graphic.graphic_node import GeneralGraphicNode, GeneralGraphicAttribute
from all_nodes.logic.class_registry import CLASS_REGISTRY as CR
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes.logic.logic_scene import LogicScene
from all_nodes import utils
from all_nodes.graphic.widgets.attribute_picker import AttributePicker
from all_nodes.graphic.widgets.class_searcher import ClassSearcher
from all_nodes.graphic.widgets.global_signaler import GlobalSignaler
from all_nodes.graphic.widgets.small_widgets import FeedbackLineEdit


GS = GlobalSignaler()

LOGGER = utils.get_logger(__name__)


# -------------------------------- CUSTOM GRAPHICS VIEW -------------------------------- #
class CustomGraphicsView(QtWidgets.QGraphicsView):
    def __init__(self):
        QtWidgets.QGraphicsView.__init__(self)
        self.setRenderHints(QtGui.QPainter.Antialiasing)
        self.setAcceptDrops(True)

        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.NoAnchor)

        self.drag_pos = None
        self.middle_pressed = False
        self.left_pressed = False

        self.feedback_lines = []
        self.feedback_line = QtWidgets.QLineEdit(parent=self)
        self.feedback_line.setReadOnly(True)
        self.feedback_line.setFont(QtGui.QFont("arial", 12))
        self.feedback_line.hide()

        self.run_btn = QtWidgets.QPushButton("Run scene", parent=self)
        self.run_btn.setFixedSize(160, 35)
        self.run_btn.setFont(QtGui.QFont("arial", 10))
        self.run_btn.setIcon(QtGui.QIcon("icons:brain.png"))

        self.reset_btn = QtWidgets.QPushButton("Reset scene", parent=self)
        self.reset_btn.setFixedSize(160, 35)
        self.reset_btn.setFont(QtGui.QFont("arial", 10))
        self.reset_btn.setIcon(QtGui.QIcon("icons:reset.png"))

        self.hourglass_animation = QtWidgets.QLabel(parent=self)
        self.hourglass_animation.setAlignment(QtCore.Qt.AlignCenter)
        ag_file = "ui:hourglass.gif"
        self.movie = QtGui.QMovie(ag_file, QtCore.QByteArray(), self)
        self.movie.setCacheMode(QtGui.QMovie.CacheAll)
        self.movie.setSpeed(100)
        self.hourglass_animation.setMovie(self.movie)
        self.movie.start()
        self.hourglass_animation.hide()

        self.class_searcher = ClassSearcher(parent=self)
        self.class_searcher.hide()
        GS.signals.class_searcher_move.connect(self.move_search_bar)

        self.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)

        # Signals connection
        self.reset_btn.clicked.connect(self.reset)
        self.run_btn.clicked.connect(self.run)

        GS.signals.execution_started.connect(self.hourglass_animation.show)
        GS.signals.execution_finished.connect(self.hourglass_animation.hide)
        GS.signals.class_scanning_finished.connect(
            lambda: self.show_feedback("Finished scanning classes", level=logging.INFO)
        )

    # RUN AND REFRESH ----------------------
    def run(self):
        """
        Run the current scene.
        """
        if self.scene():
            self.scene().run_graphic_scene()
            GS.signals.attribute_editor_global_refresh_requested.emit()

    def reset(self):
        """
        Reset the current scene.
        """
        if self.scene():
            self.scene().reset_graphic_scene()
            GS.signals.attribute_editor_global_refresh_requested.emit()

    # UTILITY ----------------------
    def show_feedback(self, message, level=logging.INFO):
        """
        Display a basic piece of feedback on screen, on top of the graphics scene.

        Args:
            message (str): message to display
            level (int): level of the message
        """
        new_feedback_line = FeedbackLineEdit(parent=self)

        effect = QtWidgets.QGraphicsOpacityEffect(new_feedback_line)
        anim = QtCore.QPropertyAnimation(effect, b"opacity", parent=self)
        anim.setStartValue(1)
        anim.setEndValue(0)
        anim.setDuration(5500)

        if level == logging.DEBUG:
            if not constants.IN_DEV:
                return
            new_feedback_line.setStyleSheet(
                "color:cyan; background-color:transparent; border:none"
            )
            message = "[DEBUG] " + message
        elif level == logging.INFO:
            new_feedback_line.setStyleSheet(
                "color:lime; background-color:transparent; border:none"
            )
            message = "[INFO] " + message
        elif level == logging.WARNING:
            new_feedback_line.setStyleSheet(
                "color:orange; background-color:transparent; border:none"
            )
            message = "[WARNING] " + message
        elif level == logging.ERROR:
            new_feedback_line.setStyleSheet(
                "color:red; background-color:transparent; border:none"
            )
            message = "[ERROR] " + message
            anim.setDuration(8000)

        new_feedback_line.show()
        new_feedback_line.setText(message)
        new_feedback_line.setGraphicsEffect(effect)
        new_feedback_line.stackUnder(self.reset_btn)
        new_feedback_line.stackUnder(self.run_btn)
        anim.start()

        new_feedback_line.setFixedSize(self.width(), 30)
        self.feedback_lines.insert(0, new_feedback_line)

        if len(self.feedback_lines) > 5:
            self.feedback_lines.pop()

        for line in self.feedback_lines:
            line.move(25, self.height() - 50 - self.feedback_lines.index(line) * 30)

    def move_search_bar(self, x, y):
        """
        Move the classes search bar to a new position

        Args:
            x (int)
            y (int)
        """
        self.class_searcher.move(self.mapFromGlobal(QtCore.QPoint(x, y)))
        self.class_searcher.reset()

    # RESIZE EVENTS ----------------------
    def resizeEvent(self, event):
        self.hourglass_animation.move(self.width() - 200, self.height() - 200)
        self.hourglass_animation.setFixedSize(200, 200)
        self.movie.setScaledSize(QtCore.QSize(200, 200))

        for line in self.feedback_lines:
            line.move(25, self.height() - 50 - self.feedback_lines.index(line) * 30)
            line.setFixedSize(self.width(), 30)

        self.reset_btn.move(self.width() - 380, self.height() - 55)
        self.run_btn.move(self.width() - 200, self.height() - 55)

        QtWidgets.QGraphicsView.resizeEvent(self, event)

    # MOUSE EVENTS ----------------------
    def mousePressEvent(self, event):
        self.middle_pressed = False

        self.drag_pos = event.pos()
        if event.button() == QtCore.Qt.MidButton:
            self.middle_pressed = True
        elif event.button() == QtCore.Qt.LeftButton:
            self.left_pressed = True
            self.class_searcher.hide()

        QtWidgets.QGraphicsView.mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        new_pos = event.pos()

        if self.middle_pressed and self.drag_pos:
            disp = new_pos - self.drag_pos
            self.drag_pos = new_pos
            transform = self.transform()
            deltaX = disp.x() / transform.m11()
            deltaY = disp.y() / transform.m22()
            self.translate(deltaX, deltaY)

        QtWidgets.QGraphicsView.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.MidButton:
            self.middle_pressed = False
            self.drag_pos = None

        QtWidgets.QGraphicsView.mouseReleaseEvent(self, event)

    def wheelEvent(self, event):
        zoom_increase = 1.45
        zoom_decrease = 0.65

        if event.delta() > 0:
            self.scale(zoom_increase, zoom_increase)
        else:
            self.scale(zoom_decrease, zoom_decrease)

    # DRAG AND DROP EVENT ----------------------
    def dragEnterEvent(self, event):
        QtWidgets.QGraphicsView.dragEnterEvent(self, event)
        event.acceptProposedAction()

    def dragMoveEvent(self, event):
        QtWidgets.QGraphicsView.dragMoveEvent(self, event)
        event.acceptProposedAction()


# -------------------------------- CUSTOM SCENE CLASS -------------------------------- #
class CustomScene(QtWidgets.QGraphicsScene):
    def __init__(self, context=None):
        QtWidgets.QGraphicsScene.__init__(self)

        # Logic scene
        self.logic_scene = context.internal_scene if context else LogicScene()
        self.filepath = context.CONTEXT_DEFINITION_FILE if context else None

        # Nodes
        self.all_graphic_nodes = set()

        # Annotations
        self.all_graphic_annotations = set()

        # Path tests
        self.testing_graphic_attr = None

        self.testing_path = ConnectorLine()
        self.testing_path.set_testing_appearance()
        self.testing_path.hide()
        self.addItem(self.testing_path)

    # SCENE SETUP ----------------------
    def drawBackground(self, painter, rect):
        pen = QtGui.QPen(QtGui.QColor(255, 230, 255, 150), 2)
        painter.setPen(pen)
        for i in range(-20_000, 20_000, 200):
            if i > rect.x() and i < rect.x() + rect.width():
                painter.drawLine(QtCore.QLine(i, -20_000, i, 20_000))
            if i > rect.y() and i < rect.y() + rect.height():
                painter.drawLine(QtCore.QLine(-20_000, i, 20_000, i))

    # ANNOTATIONS ----------------------
    def add_annotation_by_type(
        self, annotation_type: str, x: int = 0, y: int = 0
    ) -> GeneralGraphicAnnotation:
        new_annotation = GeneralGraphicAnnotation(annotation_type)
        new_annotation.setPos(x, y)
        self.all_graphic_annotations.add(new_annotation)
        self.addItem(new_annotation)
        return new_annotation

    # ADD AND DELETE NODES ----------------------
    def add_graphic_node_by_class_name(
        self, node_classname: str, x: int = 0, y: int = 0
    ) -> GeneralGraphicNode:
        """
        Create a new graphic node just by giving the class name

        Args:
            node_classname (str): class name to be instantiated
            x (int): optional, coords to place the node at
            y (int): optional, coords to place the node at

        Returns:
            GeneralGraphicNode: newly created node
        """
        all_classes = CR.get_all_classes()
        for lib in sorted(all_classes):
            for m in all_classes[lib]:
                color = all_classes[lib][m]["color"]
                for name, _ in all_classes[lib][m]["classes"]:
                    if node_classname == name:
                        new_logic_node = self.logic_scene.add_node_by_name(
                            node_classname
                        )
                        new_graph_node = GeneralGraphicNode(new_logic_node, color)
                        self.addItem(new_graph_node)
                        self.all_graphic_nodes.add(new_graph_node)
                        new_graph_node.setPos(x, y)
                        GS.signals.main_screen_feedback.emit(
                            "Created graphic node {}".format(node_classname),
                            logging.INFO,
                        )
                        return new_graph_node

        GS.signals.main_screen_feedback.emit(
            "Could not create graphic node {}".format(node_classname),
            logging.ERROR,
        )
        return None

    def add_graphic_node_from_logic_node(
        self, logic_node, x: int = 0, y: int = 0
    ) -> GeneralGraphicNode:
        """
        Create a new graphic node from an existing logic node

        Args:
            logic_node (GeneralLogicNode): logic node to be represented graphically
            x (int): optional, coords to place the node at
            y (int): optional, coords to place the node at

        Returns:
            GeneralGraphicNode: newly created node
        """
        all_classes = CR.get_all_classes()
        for lib in sorted(all_classes):
            for m in all_classes[lib]:
                color = all_classes[lib][m]["color"]
                for name, _ in all_classes[lib][m]["classes"]:
                    if logic_node.class_name == name:
                        new_graph_node = GeneralGraphicNode(logic_node, color)
                        self.addItem(new_graph_node)
                        self.all_graphic_nodes.add(new_graph_node)
                        LOGGER.info(
                            "Created graphic node from logic node {} at x:{} y:{}".format(
                                logic_node.node_name, x, y
                            )
                        )
                        new_graph_node.moveBy(x, y)
                        if not logic_node.active:
                            new_graph_node.show_deactivated()
                        return new_graph_node

    def delete_node(self, graphic_node: GeneralGraphicNode):
        """
        Delete graphic node from this scene.

        Args:
            graphic_node (GeneralGraphicNode): node to delete
        """
        self.clear_node_lines(graphic_node)
        graphic_node.clear_all_connections()
        self.all_graphic_nodes.remove(graphic_node)
        self.removeItem(graphic_node)

        self.logic_scene.remove_node_by_name(graphic_node.logic_node.node_name)

        del graphic_node

    def delete_annotation(self, graphic_annotation: GeneralGraphicAnnotation):
        self.all_graphic_annotations.remove(graphic_annotation)
        self.removeItem(graphic_annotation)
        del graphic_annotation

    # NODE CONNECTIONS/LINES ----------------------
    def connect_graphic_attrs(
        self,
        graphic_attr_1: GeneralGraphicAttribute,
        graphic_attr_2: GeneralGraphicAttribute,
        check_logic=True,
    ):
        """
        Attempt to connect two graphic attributes.

        Args:
            graphic_attr_1 (GeneralGraphicAttribute): graphic attribute to be connected
            graphic_attr_2 (GeneralGraphicAttribute): graphic attribute to be connected
            check_logic (bool): optional, whether or not the possibility of connecting should be checked
        """

        # Remove previously existing lines between these graphic_attrs
        for item in self.items():
            if item.data(0) == constants.CONNECTOR_LINE:
                if graphic_attr_1 in [
                    item.data(1),
                    item.data(2),
                ] and graphic_attr_2 in [
                    item.data(1),
                    item.data(2),
                ]:
                    self.removeItem(item)

        # Connect
        can_connect, reason = graphic_attr_1.connect_graphic_attr(
            graphic_attr_2, check_logic
        )
        if can_connect:
            GS.signals.main_screen_feedback.emit(
                reason,
                logging.DEBUG,
            )
            self.draw_valid_line(graphic_attr_1, graphic_attr_2)
        else:
            GS.signals.main_screen_feedback.emit(
                reason,
                logging.WARNING,
            )

    def draw_valid_line(self, graphic_attr_1, graphic_attr_2):
        """
        Draw a line in between to graphic attributes that we know can be connected.

        Args:
            graphic_attr_1 (GeneralGraphicAttribute)
            graphic_attr_2 (GeneralGraphicAttribute)
        """
        connector_line = ConnectorLine(graphic_attr_1, graphic_attr_2)
        self.addItem(connector_line)
        connector_line.setZValue(-1)

    def redraw_node_lines(self, node: GeneralGraphicNode):
        """
        Redraw all input and output lines that arrive to/from a node.

        Args:
            node (GeneralGraphicNode): node to redraw lines for
        """
        graphic_attrs = [
            c
            for c in self.items()
            if c.data(0) == constants.GRAPHIC_ATTRIBUTE and c.parent_node == node
        ]

        for g in graphic_attrs:
            g_1, g_2 = None, None
            for line in self.items():
                if line.data(0) != constants.CONNECTOR_LINE:
                    continue
                g_1, g_2 = line.data(1), line.data(2)

                if g in [g_1, g_2]:
                    line.repath()

    def disconnect_graphic_attrs(
        self,
        graphic_attr_1: GeneralGraphicAttribute,
        graphic_attr_2: GeneralGraphicAttribute,
    ):
        """
        Disconnect two graphic attributes.

        Args:
            graphic_attr_1 (GeneralGraphicAttribute)
            graphic_attr_2 (GeneralGraphicAttribute)
        """
        graphic_attr_1.disconnect_from(graphic_attr_2)
        GS.signals.attribute_editor_global_refresh_requested.emit()

    def clear_node_lines(self, node: GeneralGraphicNode):
        """
        Remove all input and output lines that arrive to/from a node.

        Args:
            node (GeneralGraphicNode): node to remove lines for
        """
        for line in self.items():
            g_1, g_2 = line.data(1), line.data(2)
            if g_1 is None or g_2 is None:
                continue
            if line.data(0) == constants.CONNECTOR_LINE and node in [
                g_1.parent_node,
                g_2.parent_node,
            ]:
                self.removeItem(line)

    # SAVE AND LOAD ----------------------
    def set_filepath(self, filepath):
        self.filepath = filepath

    def save_to_file(self, filepath=None):
        """
        Save this graphics scene to a file.

        Args:
            filepath (str, optional): filepath to save to. Defaults to None.
        """
        target_file = filepath
        if not filepath:
            dialog = QtWidgets.QFileDialog()
            result = dialog.getSaveFileName(
                caption="Specify target file", filter="*.yml *.ctx"
            )
            if not result[0] or not result[1]:
                return
            target_file = result[0]
        else:
            dlg = QtWidgets.QMessageBox(parent=None)
            dlg.setWindowTitle("Save scene")
            dlg.setText(
                f"<p style='white-space:pre'>Are you sure?<br>This will overwrite the scene at:<br>{filepath}"
            )
            dlg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            dlg.setIcon(QtWidgets.QMessageBox.Question)
            button = dlg.exec_()

            if button == QtWidgets.QMessageBox.No:
                return

        # Grab logic scene info and add the xy coords
        the_dict = self.logic_scene.convert_scene_to_dict()
        for g_node in self.all_graphic_nodes:
            node_name = g_node.logic_node.node_name
            for node_dict in the_dict["nodes"]:
                if node_name in node_dict:
                    node_dict[node_name]["x_pos"] = int(g_node.scenePos().x())
                    node_dict[node_name]["y_pos"] = int(g_node.scenePos().y())
                    break

        # Add annotations
        if self.all_graphic_annotations:
            the_dict["annotations"] = list()
            annotation_count = 0
            for ann in self.all_graphic_annotations:
                annotation_dict = dict()
                annotation_dict["annotation_type"] = ann.get_type()
                annotation_dict["x_pos"] = int(ann.scenePos().x())
                annotation_dict["y_pos"] = int(ann.scenePos().y())
                if ann.get_text():
                    annotation_dict["text"] = str(ann.get_text())
                the_dict["annotations"].append(annotation_dict)
                annotation_count += 1

        # Save logic scene
        self.logic_scene.save_to_file(target_file, the_dict)

    def load_from_file(self, source_file: str, create_logic_nodes=True):
        """
        Create a graphic scene from a file.

        Args:
            source_file (str): filepath of scene to load
        """
        # Set filepath
        self.set_filepath(source_file)

        # Grab the scene dict and create logic nodes
        scene_dict = dict()
        with open(source_file, "r") as file:
            scene_dict = yaml.safe_load(file)

        new_logic_nodes = self.logic_scene.all_logic_nodes
        if create_logic_nodes:
            new_logic_nodes = self.logic_scene.load_from_file(source_file)

        # Create graphic nodes
        for logic_node in new_logic_nodes:
            for node_dict in scene_dict["nodes"]:
                node_name = next(iter(node_dict))
                if logic_node.node_name == node_name:
                    self.add_graphic_node_from_logic_node(
                        logic_node,
                        node_dict[node_name]["x_pos"],
                        node_dict[node_name]["y_pos"],
                    )
                    break

        # Create annotations
        if "annotations" in scene_dict:
            for ann_dict in scene_dict["annotations"]:
                new_annotation = self.add_annotation_by_type(
                    ann_dict["annotation_type"], ann_dict["x_pos"], ann_dict["y_pos"]
                )
                new_annotation.set_text(ann_dict.get("text", ""))

        # Connections
        all_graphic_attrs = []
        for g_node in self.all_graphic_nodes:
            if g_node.logic_node in new_logic_nodes:
                for g_attribute in g_node.graphic_attributes:
                    all_graphic_attrs.append(g_attribute)

        for g_attribute in all_graphic_attrs:
            if g_attribute.connector_type != constants.OUTPUT:
                continue
            connected_logic_attributes = []
            for (
                connected_logic_attribute
            ) in g_attribute.logic_attribute.connected_attributes:
                connected_logic_attributes.append(connected_logic_attribute.dot_name)
            for logic_attr_full_name in connected_logic_attributes:
                for other_g_attribute in all_graphic_attrs:
                    if (
                        other_g_attribute.logic_attribute.dot_name
                        == logic_attr_full_name
                    ):
                        LOGGER.info(
                            "Connected graphic attributes {} -> {}".format(
                                g_attribute.logic_attribute.dot_name,
                                other_g_attribute.logic_attribute.dot_name,
                            )
                        )
                        self.connect_graphic_attrs(
                            g_attribute, other_g_attribute, check_logic=False
                        )
                        break

    # NODE-SPECIFIC ----------------------
    def rename_graphic_node(self, graphic_node: GeneralGraphicNode):
        """
        Rename a graphic node in the scene.

        Args:
            graphic_node (GeneralGraphicNode): node to be renamed

        """
        logic_node = graphic_node.logic_node
        new_name, ok = QtWidgets.QInputDialog().getText(
            None,
            "Rename node " + logic_node.node_name,
            "New name:",
        )  # TODO find how to style this

        if ok and new_name:
            if self.logic_scene.rename_node(logic_node, new_name):
                graphic_node.update_name()
                GS.signals.attribute_editor_refresh_node_requested.emit(logic_node.uuid)
                GS.signals.tab_names_refresh_requested.emit()
            else:
                if logic_node.node_name == new_name:
                    GS.signals.main_screen_feedback.emit(
                        "Cannot rename to '{}', it is the same name as now".format(
                            new_name
                        ),
                        logging.INFO,
                    )
                else:
                    if not GeneralLogicNode.name_is_valid(new_name):
                        GS.signals.main_screen_feedback.emit(
                            "Cannot rename to '{}', naming must start with uppercase and "
                            "cannot include special characters".format(new_name),
                            logging.WARNING,
                        )
                    else:
                        GS.signals.main_screen_feedback.emit(
                            "Cannot rename to '{}', "
                            "check that no other node has that name".format(new_name),
                            logging.WARNING,
                        )

    def examine_code(self, graphic_node_list: list = None):
        """
        Show the code of a node in an external editor.

        If the node is a context, show its .yml definition file as well.

        Args:
            graphic_node_list (list): node to examine code from
        """
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()

        commands = []

        if platform.system() == "Windows":
            commands.extend(["notepad", "notepad++"])
        elif platform.system() == "Linux":
            commands.extend(["gedit", "pluma"])

        for graphic_node in graphic_node_list:
            logic_node = graphic_node.logic_node

            for command in commands:
                try:
                    subprocess.Popen([command, logic_node.FILEPATH])
                    if logic_node.IS_CONTEXT:
                        subprocess.Popen([command, logic_node.CONTEXT_DEFINITION_FILE])

                    break

                except Exception as e:
                    msg = "Could not use editor '{}': {}".format(command, e)
                    LOGGER.debug(msg)

    def expand_contexts(self, graphic_node_list: list = None):
        """
        Launch signal to expand a selected context node.

        Args:
            graphic_node_list (list): nodes to be expanded
        """
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()
        for graphic_node in graphic_node_list:
            if graphic_node.logic_node.IS_CONTEXT:
                GS.signals.context_expansion_requested.emit(
                    graphic_node.logic_node.uuid
                )

    def show_log(self, graphic_node: GeneralGraphicNode):
        """
        For debugging purposes, print to screen internal information of a node.

        Args:
            graphic_node (GeneralGraphicNode): node to display info from.
        """
        print(
            "\n{}\n{}".format(
                graphic_node.logic_node.node_name,
                "-" * len(graphic_node.logic_node.node_name),
            )
        )
        pprint.pprint(graphic_node.logic_node.get_node_full_dict())

    def run_nodes(self, graphic_node_list: list = None):
        """
        Execute a graphic node.

        Args:
             graphic_node_list (list): nodes to execute
        """
        # Check nodes to execute
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()

        if not graphic_node_list:
            GS.signals.main_screen_feedback.emit(
                "No nodes selected, nothing to execute", logging.WARNING
            )
            return

        # Execution
        GS.signals.main_screen_feedback.emit(
            "Running only selected node(s)", logging.INFO
        )
        GS.signals.execution_started.emit()

        self.logic_scene.run_list_of_nodes(
            [graphic_node.logic_node for graphic_node in graphic_node_list]
        )

    def reset_nodes(self, graphic_node_list: list = None):
        """
        Reset appearance of a graphic node and then also reset its
        internal logic node.

        Args:
             graphic_node_list (list): nodes to reset
        """
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()
        for graphic_node in graphic_node_list:
            graphic_node.reset()
            graphic_node.logic_node.reset()

        GS.signals.main_screen_feedback.emit("Resetting selected node(s)", logging.INFO)
        GS.signals.attribute_editor_global_refresh_requested.emit()

    def soft_reset_nodes(self, graphic_node_list: list = None):
        """
        Reset appearance of a graphic node and then also soft-reset its
        internal logic node.

        Args:
            graphic_node_list (list): nodes to soft-reset
        """
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()
        for graphic_node in graphic_node_list:
            graphic_node.reset()
            graphic_node.logic_node.soft_reset()

        GS.signals.main_screen_feedback.emit(
            "Soft-resetting selected node(s)", logging.INFO
        )
        GS.signals.attribute_editor_global_refresh_requested.emit()

    def toggle_activated_nodes(self, graphic_node_list: list = None):
        """
        Toggle whether or not a graphic node is activated.

        Args:
            graphic_node_list (list): nodes to toggle activation of
        """
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()
        for graphic_node in graphic_node_list:
            graphic_node.toggle_activated()

        # GS.signals.attribute_editor_global_refresh_requested.emit()  # TODO consider this in attribute editor?

    def add_attr_to_node(self, graphic_node: GeneralGraphicNode):
        a_picker = AttributePicker()
        if a_picker.get_results():
            graphic_node.add_single_graphic_attribute(*a_picker.get_results())

    def export_nodes_code(self, graphic_node_list: list):
        if not graphic_node_list:
            graphic_node_list = self.selected_nodes()

        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

        py_contents = (
            f"# ------ File written by 'all_nodes' {timestamp} ------ # \n\n"
            "from all_nodes.constants import InputsGUI, PreviewsGUI \n"
            "from all_nodes.logic.logic_node import GeneralLogicNode \n"
            "from all_nodes import utils \n\n\n"
            "LOGGER = utils.get_logger(__name__)\n\n\n"
        )

        dialog = QtWidgets.QFileDialog()
        result = dialog.getExistingDirectory(
            caption="Specify target file", filter="*.yml *.ctx"
        )
        if not result:
            return

        out_path = os.path.join(result, f"all_nodes_{timestamp}.py")

        with open(out_path, "w") as f:
            f.write(py_contents)

        for g_node in graphic_node_list:
            code = g_node.get_as_code()
            code = re.sub(
                "<class '(?P<datatype>[a-zA-Z_.0-9]+)'>", "\g<datatype>", code
            )
            code = re.sub(
                "<(?P<guitype>[a-zA-Z0-9._]+): '[a-zA-Z ]+'+>",
                "\g<guitype>",
                code,
            )
            with open(out_path, "a") as f:
                f.write(code)
                f.write("\n\n\n")

        try:
            subprocess.Popen([sys.executable, "-m", "ruff", "format", out_path])
        except Exception as e:
            msg = "Could not use ruff: {}".format(e)
            LOGGER.info(msg)

        LOGGER.info(f"Exported code to {out_path}")
        GS.signals.main_screen_feedback.emit(
            f"Exported code to {out_path}", logging.INFO
        )

    def deselect_all(self):
        """
        Deselect all graphic nodes.
        """
        for n in self.all_graphic_nodes:
            n.setSelected(False)

    # ANNOTATION SPECIFIC ----------------------
    def bring_annotation_to_front(self, annotation_list: list):
        if not annotation_list:
            annotation_list = self.selected_annotations()

        for annotation in annotation_list:
            annotation.setZValue(500)

    def move_annotation_backward(self, annotation_list: list):
        if not annotation_list:
            annotation_list = self.selected_annotations()

        for annotation in annotation_list:
            annotation.setZValue(-500)

    # SCENE EXECUTION ----------------------
    def reset_all_graphic_nodes(self):
        """
        Reset all graphic nodes.
        """
        utils.print_separator("Resetting all graphic nodes")
        for g_node in self.all_graphic_nodes:
            g_node.reset()

    def show_result_on_nodes(self):
        """
        Display the internal result of each graphic node.
        """
        for g_node in self.all_graphic_nodes:
            g_node.show_result()

    def reset_graphic_scene(self):
        """
        Reset the appearance of all the nodes in this graphic scene.
        """
        self.reset_all_graphic_nodes()
        self.logic_scene.reset_all_nodes()
        GS.signals.attribute_editor_global_refresh_requested.emit()

    def run_graphic_scene(self):
        """
        Run all the nodes in this graphic scene.
        """
        GS.signals.execution_started.emit()
        self.reset_all_graphic_nodes()
        self.logic_scene.run_all_nodes()

    # UTILITY ----------------------
    def selected_nodes(self) -> list:
        """
        Gather all selected nodes.

        Returns:
            list: with all selected nodes.
        """
        sel_nodes = []
        for n in self.all_graphic_nodes:
            if n.isSelected():
                sel_nodes.append(n)
        return sel_nodes

    def selected_annotations(self) -> list:
        sel_annotations = []
        for a in self.all_graphic_annotations:
            if a.isSelected():
                sel_annotations.append(a)
        return sel_annotations

    def fit_in_view(self):
        """
        If selected nodes, fit all of them in the view. Otherwise, fit all nodes.
        """
        self.parent().resetMatrix()
        nodes = list(self.all_graphic_nodes) + list(self.all_graphic_annotations)
        if self.selected_nodes() or self.selected_annotations():
            nodes = list(self.selected_nodes()) + list(self.selected_annotations())
        if not nodes:
            return
        rect = nodes[0].sceneBoundingRect()
        for n in nodes:
            rect = rect.united(n.sceneBoundingRect())

        self.parent().fitInView(
            rect + QtCore.QMarginsF(20.0, 20.0, 20.0, 20.0), QtCore.Qt.KeepAspectRatio
        )

    # CONTEXT EVENTS ----------------------
    def contextMenuEvent(self, event):
        selection_rect = QtCore.QRect(
            event.scenePos().x() - 1,
            event.scenePos().y() - 1,
            1,
            1,
        )
        test_items = self.items(selection_rect)
        if test_items:
            for item in test_items:
                if item and item.data(0) == constants.GRAPHIC_NODE:
                    self.menu_node(event, item)
                    break
                elif item and item.data(0) == constants.GRAPHIC_ANNOTATION:
                    self.menu_annotation(event, item)
                    break
        else:
            self.menu_scene(event)

        QtWidgets.QGraphicsScene.contextMenuEvent(self, event)

    def menu_node(self, event, node):
        menu = QtWidgets.QMenu()  # TODO add tooltips

        # Common actions
        rename_action = menu.addAction(" Rename this node")
        rename_action.setIcon(QtGui.QIcon("icons:rename.png"))
        rename_action.triggered.connect(lambda: self.rename_graphic_node(node))

        examine_code_action = menu.addAction(" Examine code")
        examine_code_action.setIcon(QtGui.QIcon("icons:examine.png"))
        examine_code_action.triggered.connect(lambda: self.examine_code([node]))

        # Execution-related  actions
        menu.addSeparator()

        run_nodes_action = menu.addAction(" Run only this node")
        run_nodes_action.setIcon(QtGui.QIcon("icons:run_one.svg"))
        run_nodes_action.triggered.connect(lambda: self.run_nodes([node]))

        reset_nodes_action = menu.addAction(" Reset only this node")
        reset_nodes_action.setIcon(QtGui.QIcon("icons:reset.png"))
        reset_nodes_action.triggered.connect(lambda: self.reset_nodes([node]))

        deactivate_single_node_action = menu.addAction(
            " Activate/deactivate only this node"
        )
        deactivate_single_node_action.setIcon(QtGui.QIcon("icons:close.svg"))
        deactivate_single_node_action.triggered.connect(
            lambda: self.toggle_activated_nodes([node])
        )

        # Context-specific
        if node.logic_node.IS_CONTEXT:
            menu.addSeparator()
            expand_context_action = menu.addAction(" Expand context")
            expand_context_action.setIcon(QtGui.QIcon("icons:cubes.svg"))
            expand_context_action.triggered.connect(
                lambda: self.expand_contexts([node])
            )

        # Add attributes
        menu.addSeparator()
        add_attr_to_node_action = menu.addAction(" Add new attribute to this node")
        add_attr_to_node_action.setIcon(QtGui.QIcon("icons:plus.svg"))
        add_attr_to_node_action.triggered.connect(lambda: self.add_attr_to_node(node))

        # Export code
        export_nodes_code_action = menu.addAction(" Export this node's code")
        export_nodes_code_action.setIcon(QtGui.QIcon("icons:wand.svg"))
        export_nodes_code_action.triggered.connect(
            lambda: self.export_nodes_code([node])
        )

        # Debug
        menu.addSeparator()
        soft_reset_nodes_action = menu.addAction(" [DEBUG] Soft-reset only this node")
        soft_reset_nodes_action.setIcon(QtGui.QIcon("icons:reset.png"))
        soft_reset_nodes_action.triggered.connect(lambda: self.soft_reset_nodes([node]))

        # DEV-specific
        if constants.IN_DEV:
            menu.addSeparator()
            print_node_log_action = menu.addAction(" Print node log")
            print_node_log_action.setIcon(QtGui.QIcon("icons:nodes_2.png"))
            print_node_log_action.triggered.connect(lambda: self.show_log(node))

        # Style
        f = QtCore.QFile(r"ui:stylesheet.qss")  # TODO not ideal, maybe a reduced qss?
        with open(f.fileName(), "r") as s:
            menu.setStyleSheet(s.read())

        # Exec
        menu.exec_(event.screenPos(), parent=self)

    def menu_annotation(self, event, node):
        menu = QtWidgets.QMenu()  # TODO add tooltips

        bring_forward_action = menu.addAction(" Bring annotation to front")
        bring_forward_action.setIcon(QtGui.QIcon("icons:front.svg"))
        bring_forward_action.triggered.connect(
            lambda: self.bring_annotation_to_front([node])
        )

        move_backward_action = menu.addAction(" Move annotation back")
        move_backward_action.setIcon(QtGui.QIcon("icons:back.svg"))
        move_backward_action.triggered.connect(
            lambda: self.move_annotation_backward([node])
        )

        # Style
        f = QtCore.QFile(r"ui:stylesheet.qss")  # TODO not ideal, maybe a reduced qss?
        with open(f.fileName(), "r") as s:
            menu.setStyleSheet(s.read())

        # Exec
        menu.exec_(event.screenPos(), parent=self)

    def menu_scene(self, event):
        menu = QtWidgets.QMenu()  # TODO add tooltips

        # Common actions
        run_action = menu.addAction(" Run this scene")
        run_action.setIcon(QtGui.QIcon("icons:brain.png"))
        run_action.triggered.connect(self.run_graphic_scene)

        # Saving
        if self.filepath:
            menu.addSeparator()
            save_action = menu.addAction(
                f" Save changes to {os.path.basename(self.filepath)}"
            )
            save_action.setIcon(QtGui.QIcon("icons:save.svg"))
            save_action.triggered.connect(lambda: self.save_to_file(self.filepath))
        else:
            save_action = menu.addAction(" Save scene as")
            save_action.setIcon(QtGui.QIcon("icons:save_as.svg"))
            save_action.triggered.connect(self.save_to_file)

        # Style
        f = QtCore.QFile(r"ui:stylesheet.qss")  # TODO not ideal, maybe a reduced qss?
        with open(f.fileName(), "r") as s:
            menu.setStyleSheet(s.read())

        # Exec
        menu.exec_(event.screenPos(), parent=self)

    # KEYBOARD EVENTS ----------------------
    def event(self, event: QtWidgets.QGraphicsScene.event):
        # Only use keyboard events
        if not (event.type() == QtCore.QEvent.KeyPress):
            return QtWidgets.QGraphicsScene.event(self, event)

        # See that we are not working on a node widget
        if self.focusItem():  # Means we are editing a widget in some input node
            return QtWidgets.QGraphicsScene.event(self, event)

        modifiers = QtWidgets.QApplication.keyboardModifiers()

        # --- Examine code
        if event.key() == QtCore.Qt.Key_E and not modifiers:
            self.examine_code(self.selected_nodes())

        # --- Execution
        elif event.key() == QtCore.Qt.Key_Return and not modifiers:
            self.run_nodes(self.selected_nodes())

        elif event.key() == QtCore.Qt.Key_R and not modifiers:
            self.reset_nodes(self.selected_nodes())

        # --- Toggle activation
        elif event.key() == QtCore.Qt.Key_D and not modifiers:
            self.toggle_activated_nodes(self.selected_nodes())

        # --- Context expansion
        elif (
            event.key() == QtCore.Qt.Key_Return
            and modifiers == QtCore.Qt.ControlModifier
        ):
            self.expand_contexts(self.selected_nodes())

        # --- Export code
        elif event.key() == QtCore.Qt.Key_E and modifiers == QtCore.Qt.ControlModifier:
            self.export_nodes_code(self.selected_nodes())

        # --- Soft-reset
        elif event.key() == QtCore.Qt.Key_S:
            self.soft_reset_nodes(self.selected_nodes())

        # --- Delete
        elif event.key() == QtCore.Qt.Key_Delete:
            for n in self.selected_nodes():
                GS.signals.attribute_editor_remove_node_requested.emit(
                    n.logic_node.uuid
                )
                self.delete_node(n)

            for a in self.selected_annotations():
                self.delete_annotation(a)

        # --- Fit in view
        elif event.key() == QtCore.Qt.Key_F:
            self.fit_in_view()

        # --- Class search
        elif event.key() == QtCore.Qt.Key_Tab and not modifiers:
            GS.signals.class_searcher_move.emit(
                QtGui.QCursor.pos().x(), QtGui.QCursor.pos().y()
            )

        return QtWidgets.QGraphicsScene.event(self, event)

    # MOUSE EVENTS ----------------------
    def mousePressEvent(self, event):
        QtWidgets.QGraphicsScene.mousePressEvent(
            self, event
        )  # TODO check if these have to go at the bottom and with return

        if not event.button() == QtCore.Qt.LeftButton:
            return

        test_item = None
        selection_rect = QtCore.QRect(
            event.scenePos().x() - 3 * constants.PLUG_RADIUS,
            event.scenePos().y() - constants.PLUG_RADIUS,
            6 * constants.PLUG_RADIUS,
            2 * constants.PLUG_RADIUS,
        )
        test_items = self.items(selection_rect)
        for item in test_items:
            if item and item.data(0) == constants.PLUG:
                test_item = item
                break

        if test_item:
            self.testing_graphic_attr = test_item.parentItem()

            if (
                self.testing_graphic_attr.has_input_connected()
            ):  # It is input and had connection
                connected_out_graphic_attr = next(
                    iter(self.testing_graphic_attr.connected_graphic_attrs)
                )
                self.disconnect_graphic_attrs(
                    connected_out_graphic_attr, self.testing_graphic_attr
                )
                for item in self.items():
                    if item.data(0) == constants.CONNECTOR_LINE:
                        if self.testing_graphic_attr in [item.data(1), item.data(2)]:
                            self.removeItem(item)
                self.testing_graphic_attr = connected_out_graphic_attr

    def mouseDoubleClickEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        QtWidgets.QGraphicsScene.mouseDoubleClickEvent(self, event)
        selection_rect = QtCore.QRect(
            event.scenePos().x() - 1,
            event.scenePos().y() - 1,
            1,
            1,
        )
        test_items = self.items(selection_rect)
        for item in test_items:
            if item and item.data(0) == constants.GRAPHIC_NODE:
                GS.signals.attribute_editor_node_addition_requested.emit(
                    item.logic_node.uuid
                )
                break

    def mouseMoveEvent(self, event):
        QtWidgets.QGraphicsScene.mouseMoveEvent(self, event)

        if self.testing_graphic_attr:
            self.testing_path.show()

            test_origin = self.testing_graphic_attr.plug_coords()

            p1 = test_origin
            p2 = event.scenePos()

            new_path = ConnectorLine.calculate_path(p1, p2)

            if new_path.length() > 7000:
                self.testing_path.hide()
            else:
                self.testing_path.show()

            self.testing_path.set_new_path(new_path)

    def mouseReleaseEvent(self, event):
        QtWidgets.QGraphicsScene.mouseReleaseEvent(self, event)

        if not event.button() == QtCore.Qt.LeftButton:
            return

        event_x = event.scenePos().x()
        event_y = event.scenePos().y()

        test_item = None
        selection_rect = QtCore.QRect(
            event_x - 3 * constants.PLUG_RADIUS,
            event_y - constants.PLUG_RADIUS,
            6 * constants.PLUG_RADIUS,
            2 * constants.PLUG_RADIUS,
        )
        test_items = self.items(selection_rect)
        for item in test_items:
            if item and item.data(0) == constants.PLUG:
                test_item = item
                break

        if self.testing_graphic_attr and test_item:
            graphic_attr_1 = self.testing_graphic_attr
            graphic_attr_2 = test_item.parentItem()
            if graphic_attr_1.parent_node != graphic_attr_2.parent_node:
                if graphic_attr_2.connector_type == constants.INPUT:
                    graphic_attr_2.disconnect_input()
                    for line in self.items():
                        if line.data(
                            0
                        ) == constants.CONNECTOR_LINE and graphic_attr_2 in [
                            line.data(1),
                            line.data(2),
                        ]:
                            self.removeItem(line)
                self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)
            else:
                GS.signals.main_screen_feedback.emit(
                    "Those two attributes belong to the same node!",
                    logging.WARNING,
                )
        elif self.testing_graphic_attr and test_item is None:
            graphic_attr_1 = self.testing_graphic_attr
            if graphic_attr_1.connector_type == constants.INPUT:
                new_g_node = None
                if graphic_attr_1.logic_attribute.data_type is str:
                    new_g_node = self.add_graphic_node_by_class_name(
                        "StrInput", event_x, event_y
                    )
                    graphic_attr_2 = new_g_node["out_str"]
                    self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)
                elif graphic_attr_1.logic_attribute.data_type is list:
                    new_g_node = self.add_graphic_node_by_class_name(
                        "ListInput", event_x, event_y
                    )
                    graphic_attr_2 = new_g_node["out_list"]
                    self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)
                elif graphic_attr_1.logic_attribute.data_type is dict:
                    new_g_node = self.add_graphic_node_by_class_name(
                        "DictInput", event_x, event_y
                    )
                    graphic_attr_2 = new_g_node["out_dict"]
                    self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)
                elif graphic_attr_1.logic_attribute.data_type is int:
                    new_g_node = self.add_graphic_node_by_class_name(
                        "IntInput", event_x, event_y
                    )
                    graphic_attr_2 = new_g_node["out_int"]
                    self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)
                elif graphic_attr_1.logic_attribute.data_type is float:
                    new_g_node = self.add_graphic_node_by_class_name(
                        "FloatInput", event_x, event_y
                    )
                    graphic_attr_2 = new_g_node["out_float"]
                    self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)
                elif graphic_attr_1.logic_attribute.data_type is bool:
                    new_g_node = self.add_graphic_node_by_class_name(
                        "BoolInput", event_x, event_y
                    )
                    graphic_attr_2 = new_g_node["out_bool"]
                    self.connect_graphic_attrs(graphic_attr_1, graphic_attr_2)

                if new_g_node:
                    new_g_node.moveBy(
                        event_x
                        - graphic_attr_2.plug_coords().x()
                        + 3 * constants.PLUG_RADIUS,
                        event_y - graphic_attr_2.plug_coords().y(),
                    )

        self.testing_graphic_attr = None
        self.testing_path.hide()

        GS.signals.attribute_editor_global_refresh_requested.emit()

    # DRAG AND DROP EVENT ----------------------
    def dragEnterEvent(self, event):
        QtWidgets.QGraphicsScene.dragEnterEvent(self, event)
        event.acceptProposedAction()

    def dragMoveEvent(self, event):
        QtWidgets.QGraphicsScene.dragMoveEvent(self, event)
        event.acceptProposedAction()

    def dragLeaveEvent(self, event):
        QtWidgets.QGraphicsScene.dragLeaveEvent(self, event)
        event.acceptProposedAction()

    def dropEvent(self, event):
        # TODO make sure this can only happen after GS emits signal of all classes scanned
        event_data = event.mimeData()
        if event_data.hasUrls():
            event_urls = event_data.urls()
            for url in event_urls:
                if os.path.isfile(url.toLocalFile()):
                    self.load_from_file(url.toLocalFile())
        else:
            GS.signals.dropped_node.emit(
                self.parent().mapFromScene(event.scenePos()) - QtCore.QPoint(20, 20)
            )
        QtWidgets.QGraphicsScene.dropEvent(self, event)
        event.acceptProposedAction()


# -------------------------------- HELPER CLASES -------------------------------- #
class ConnectorLine(QtWidgets.QGraphicsPathItem):
    def __init__(self, graphic_attr_1=None, graphic_attr_2=None):
        QtWidgets.QGraphicsPathItem.__init__(self)

        # Attributes to connect
        self.graphic_attr_1 = None
        self.graphic_attr_2 = None

        # Start and end
        self.p1 = QtCore.QPointF(0.0, 0.0)
        self.p2 = QtCore.QPointF(0.0, 0.0)

        # Items
        self.glow = QtWidgets.QGraphicsPathItem(parent=self)
        self.arrow_glow = QtWidgets.QGraphicsPathItem(parent=self)
        self.white_line = QtWidgets.QGraphicsPathItem(parent=self)
        self.arrow = QtWidgets.QGraphicsPathItem(parent=self)

        # Main path ---------------------
        if graphic_attr_1 is not None and graphic_attr_2 is not None:
            if graphic_attr_1.connector_type == constants.OUTPUT:
                self.graphic_attr_1 = graphic_attr_1
                self.graphic_attr_2 = graphic_attr_2
            else:
                self.graphic_attr_1 = graphic_attr_2
                self.graphic_attr_2 = graphic_attr_1

        self.repath()

        # Glow
        if constants.GLOW_EFFECTS:
            self.glow.setPen(constants.LINE_GLOW_PEN)
            blur = QtWidgets.QGraphicsBlurEffect()
            blur.setBlurRadius(constants.LINE_GLOW_PEN.width() * 1.5)
            self.glow.setGraphicsEffect(blur)
        else:
            self.glow.hide()

        # White line
        self.white_line.setPen(constants.VALID_LINE_PEN)

        # Arrow ---------------------
        arrow_path = QtGui.QPainterPath()
        arrow_polygon = QtGui.QPolygon(
            [
                QtCore.QPoint(constants.PLUG_RADIUS * 1.3, 0),
                QtCore.QPoint(
                    -constants.PLUG_RADIUS * 1.3, constants.PLUG_RADIUS * 1.3
                ),
                QtCore.QPoint(-constants.PLUG_RADIUS * 0.9, 0),
                QtCore.QPoint(
                    -constants.PLUG_RADIUS * 1.3, -constants.PLUG_RADIUS * 1.3
                ),
            ]
        )
        arrow_path.addPolygon(arrow_polygon)
        arrow_path.closeSubpath()

        # Arrow glow
        if constants.GLOW_EFFECTS:
            self.arrow_glow.setPath(arrow_path)
            self.arrow_glow.setPen(constants.LINE_GLOW_PEN)
            blur = QtWidgets.QGraphicsBlurEffect()
            blur.setBlurRadius(constants.LINE_GLOW_PEN.width() * 1.5)
            self.arrow_glow.setGraphicsEffect(blur)
        else:
            self.arrow_glow.hide()

        # Arrow
        self.arrow.setPath(arrow_path)
        self.arrow.setPen(QtCore.Qt.NoPen)
        self.arrow.setBrush(QtGui.QBrush(constants.VALID_LINE_PEN.color()))

        self.rotate_arrows()

        # Data ---------------------
        self.setData(0, constants.CONNECTOR_LINE)
        # TODO split this data into attrs
        self.setData(1, graphic_attr_1)
        self.setData(2, graphic_attr_2)

    @staticmethod
    def calculate_path(p1, p2):
        new_path = QtGui.QPainterPath(p1)
        if constants.CONNECTOR_LINE_GEO == constants.STRAIGHT_LINES:
            new_path.lineTo(p2)
        elif constants.CONNECTOR_LINE_GEO == constants.STEPPED_LINES:
            s_p1 = QtCore.QPointF(((p2.x() + p1.x()) / 2), p1.y())
            s_p2 = QtCore.QPointF(((p2.x() + p1.x()) / 2), p2.y())
            new_path.lineTo(s_p1)
            new_path.lineTo(s_p2)
            new_path.lineTo(p2)
        elif constants.CONNECTOR_LINE_GEO == constants.SPLINE_LINES:
            c_p1 = QtCore.QPoint(((p2.x() + p1.x()) / 2) + 10, p1.y())
            c_p2 = QtCore.QPoint(((p2.x() + p1.x()) / 2) - 10, p2.y())
            new_path.cubicTo(c_p1, c_p2, p2)

        return new_path

    def repath(self):
        if self.graphic_attr_1 is None or self.graphic_attr_2 is None:
            return

        p1, p2 = (
            self.graphic_attr_1.plug_coords(),
            self.graphic_attr_2.plug_coords(),
        )

        self.set_new_path(self.calculate_path(p1, p2))
        self.rotate_arrows()

    def rotate_arrows(self):
        if self.graphic_attr_1 and self.graphic_attr_2:
            p1, p2 = self.path().pointAtPercent(0.0), self.path().pointAtPercent(1.0)
            self.arrow_glow.setPos((p2.x() + p1.x()) / 2, (p2.y() + p1.y()) / 2)
            self.arrow.setPos((p2.x() + p1.x()) / 2, (p2.y() + p1.y()) / 2)
            ref_1 = self.path().pointAtPercent(0.49)
            ref_2 = self.path().pointAtPercent(0.51)
            radians = math.atan2(ref_2.y() - ref_1.y(), ref_2.x() - ref_1.x())
            angle = math.degrees(radians)

            self.arrow_glow.setRotation(angle)
            self.arrow.setRotation(angle)

    def set_testing_appearance(self):
        self.white_line.setPen(constants.TEST_LINE_PEN)
        self.arrow_glow.hide()
        self.arrow.hide()

    def set_new_path(self, new_path):
        self.setPath(new_path)
        self.white_line.setPath(new_path)
        self.glow.setPath(new_path)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import unittest

from all_nodes.logic.class_registry import CLASS_REGISTRY as CR
from all_nodes.logic.logic_scene import LogicScene
from all_nodes import utils


# -------------------------------- TESTS -------------------------------- #
class ClassRegistryTest(unittest.TestCase):
    def test_class_registry(self):
        """
        Basic test of the class registry
        """
        utils.print_test_header("test_class_registry")

        CR.flush()
        assert CR._all_classes is None
        CR.get_all_classes()
        assert len(CR._all_classes.keys()) > 0

        CR.flush()
        assert CR._all_classes is None

    def test_class_registry_scene(self):
        """
        Test that the class registry scanning is launched when adding nodes to a scene
        """
        utils.print_test_header("test_class_registry_scene")

        CR.flush()
        assert CR._all_classes is None

        logic_scene = LogicScene()
        logic_scene.add_node_by_name("JsonToDict")

        assert len(CR._all_classes.keys()) > 2
        assert len(CR._all_classes.keys()) < 50

        CR.flush()
        assert CR._all_classes is None



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class GetDictKey(GeneralLogicNode):
    NICE_NAME = "Get dict key"
    HELP = "Get given key from a dictionary"

    INPUTS_DICT = {
        "in_dict": {"type": dict},
        "key": {"type": str},
    }

    OUTPUTS_DICT = {"out": {"type": object}}

    def run(self):
        in_dict = self.get_attribute_value("in_dict")
        key = self.get_attribute_value("key")
        if key not in in_dict:
            self.fail("Key '{}' not present in input dict {}".format(key, in_dict))
            return
        self.set_output("out", in_dict[key])


class SetDictKey(GeneralLogicNode):
    NICE_NAME = "Set dict key"
    HELP = "Set a key in a dictionary"

    INPUTS_DICT = {
        "in_dict": {"type": dict},
        "key": {"type": str},
        "new_value": {"type": object},
    }

    OUTPUTS_DICT = {"out_dict": {"type": dict}}

    def run(self):
        in_dict = self.get_attribute_value("in_dict")
        key = self.get_attribute_value("key")
        new_value = self.get_attribute_value("new_value")

        in_dict[key] = new_value
        self.set_output("out_dict", in_dict)



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class RegexMatch(GeneralLogicNode):
    INPUTS_DICT = {
        "in_str": {"type": str},
        "pattern": {"type": str},
    }

    OUTPUTS_DICT = {"match": {"type": bool}}

    def run(self):
        import re

        in_str = self.get_attribute_value("in_str")
        pattern = self.get_attribute_value("pattern")

        match = re.match(pattern, in_str)
        self.set_output("match", match is not None)



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from enum import Enum
import logging
import os

from PySide2 import QtCore
from PySide2 import QtGui


# -------------------------------- GRAPHIC -------------------------------- #
# Overall style
GLOW_EFFECTS = True
SHOW_GRID = True
IN_SCREEN_ERRORS = True

# Line style
CONNECTOR_LINE_WIDTH = 3
TEST_LINE_PEN = QtGui.QPen(QtCore.Qt.white, CONNECTOR_LINE_WIDTH, QtCore.Qt.DashLine)
LINE_GLOW_PEN = QtGui.QPen(QtCore.Qt.magenta, CONNECTOR_LINE_WIDTH * 2)
VALID_LINE_PEN = QtGui.QPen(QtCore.Qt.white, CONNECTOR_LINE_WIDTH)
STRAIGHT_LINES = 0
STEPPED_LINES = 1
SPLINE_LINES = 2
CONNECTOR_LINE_GEO = SPLINE_LINES

# Node style
STRIPE_HEIGHT = 20
CHAMFER_RADIUS = STRIPE_HEIGHT / 2
HEADER_HEIGHT = int(STRIPE_HEIGHT * 1.5)
NODE_CONTOUR_THICKNESS = 1.0
NODE_SELECTED_PEN = QtGui.QPen(QtCore.Qt.white, 2.5)
NODE_SELECTED_GLOW_THICKNESS = NODE_SELECTED_PEN.width() * 2.5
NODE_ERROR_PEN = QtGui.QPen(QtCore.Qt.red, 2, QtCore.Qt.DashDotDotLine)
NODE_DEACTIVATED_PEN = QtGui.QPen(QtGui.QColor(250, 0, 0, 235), 10)
NODE_ERROR_BRUSH = QtGui.QBrush(QtCore.Qt.red, QtCore.Qt.BDiagPattern)
NODE_FAILED_PEN = QtGui.QPen(QtGui.QColor("orange"), 2, QtCore.Qt.DashDotDotLine)
NODE_FAILED_BRUSH = QtGui.QBrush(QtGui.QColor("orange"), QtCore.Qt.BDiagPattern)
NODE_FONT = "arial"
DEFAULT_NODE_COLOR = "#4D004C"  # magenta

# Plug style
PLUG_RADIUS = int(0.3 * STRIPE_HEIGHT)
CONNECTOR_USED_PEN = QtGui.QPen(QtCore.Qt.white, 3)

# Naming
GRAPHIC_NODE = "GRAPHIC_NODE"
GRAPHIC_ATTRIBUTE = "GRAPHIC_ATTRIBUTE"

PLUG = "PLUG"
CONNECTOR_LINE = "CONNECTOR_LINE"

GRAPHIC_ANNOTATION = "GRAPHIC_ANNOTATION"


# -------------------------------- LOGIC -------------------------------- #
# Naming
INPUT = "INPUT"
OUTPUT = "OUTPUT"

INTERNAL = "INTERNAL"

START = "START"
COMPLETED = "COMPLETED"

# Node statuses
NOT_RUN = "NOT_RUN"
EXECUTING = "EXECUTING"
IN_LOOP = "IN_LOOP"
SKIPPED = "SKIPPED"
SUCCESSFUL = "SUCCESSFUL"
FAILED = "FAILED"
ERROR = "ERROR"


# -------------------------------- GUI INPUT / PREVIEW TYPES -------------------------------- #
class InputsGUI(Enum):
    STR_INPUT = "String input"
    MULTILINE_STR_INPUT = "Multiline string input"
    INT_INPUT = "Integer input"
    FLOAT_INPUT = "Float input"
    BOOL_INPUT = "Boolean input"
    OPTION_INPUT = "Option input"
    TUPLE_INPUT = "Tuple input"
    DICT_INPUT = "Dictionary input"
    LIST_INPUT = "List input"


class PreviewsGUI(Enum):
    STR_PREVIEW = "String preview"
    MULTILINE_STR_PREVIEW = "Multiline string preview"
    DICT_PREVIEW = "Dictionary preview"
    IMAGE_PREVIEW = "Image preview"


# -------------------------------- LOGGING PREFS -------------------------------- #
LOGGING_LEVEL = logging.INFO
CONSOLE_LOG_FORMATTER = logging.Formatter(
    "%(asctime)s.%(msecs)03d %(levelname)8s- %(message)s (%(funcName)s) %(filename)s:%(lineno)d",
    datefmt="%Y%m%d %H:%M:%S",
)


# -------------------------------- DEV -------------------------------- #
IN_DEV = os.getenv("IN_DEV", False)
if IN_DEV:
    LOGGING_LEVEL = logging.DEBUG
    CONSOLE_LOG_FORMATTER = logging.Formatter(
        "%(levelname)8s- %(message)s (%(funcName)s) %(filename)s:%(lineno)d"
    )

# -------------------------------- ANALYTICS -------------------------------- #
DB_ENV = os.getenv("DB_ENV", "PROD")
if IN_DEV:
    DB_ENV = "DEV"



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class ListToStr(GeneralLogicNode):
    INPUTS_DICT = {
        "in_list": {"type": list},
        "join_character": {"type": str, "optional": True},
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        in_list = [str(elem) for elem in self.get_attribute_value("in_list")]
        join_character = self.get_attribute_value("join_character") or " "
        self.set_output("out_str", join_character.join(in_list))


class IntToStr(GeneralLogicNode):
    INPUTS_DICT = {
        "in_int": {"type": int},
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        in_int = self.get_attribute_value("in_int")
        self.set_output("out_str", str(in_int))


class ConcatStr(GeneralLogicNode):
    INPUTS_DICT = {
        "in_str_0": {"type": str},
        "in_str_1": {"type": str},
        "in_str_2": {"type": str, "optional": True},
        "in_str_3": {"type": str, "optional": True},
        "in_str_4": {"type": str, "optional": True},
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        out_str = ""
        for i in range(5):
            attr_name = "in_str_{}".format(i)
            val = self.get_attribute_value(attr_name)
            if val is not None:
                out_str += val
        self.set_output("out_str", out_str)


class MultiToStr(GeneralLogicNode):
    INPUTS_DICT = {
        "in_int": {"type": int, "optional": True},
        "in_float": {"type": float, "optional": True},
        "in_bool": {"type": bool, "optional": True},
        "in_dict": {"type": dict, "optional": True},
        "in_object": {"type": object, "optional": True},
    }

    OUTPUTS_DICT = {
        "[NOTHING]": {"type": object, "optional": True},
        "out_int_to_str": {"type": str, "optional": True},
        "out_float_to_str": {"type": str, "optional": True},
        "out_bool_to_str": {"type": str, "optional": True},
        "out_dict_to_str": {"type": str, "optional": True},
        "out_object_to_str": {"type": str, "optional": True},
    }

    def run(self):
        in_int = self.get_attribute_value("in_int")
        if in_int is not None:
            self.set_output("out_int_to_str", str(in_int))

        in_float = self.get_attribute_value("in_float")
        if in_float is not None:
            self.set_output("out_float_to_str", str(in_float))

        in_bool = self.get_attribute_value("in_bool")
        if in_bool is not None:
            self.set_output("out_bool_to_str", str(in_bool))

        in_dict = self.get_attribute_value("in_dict")
        if in_dict is not None:
            self.set_output("out_dict_to_str", str(in_dict))

        in_object = self.get_attribute_value("in_object")
        if in_object is not None:
            self.set_output("out_object_to_str", str(in_object))



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.constants import InputsGUI
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class ImageFileExtensionSelect(GeneralLogicNode):
    NICE_NAME = "Image file extension select"

    INTERNALS_DICT = {
        "internal_str": {
            "type": str,
            "gui_type": InputsGUI.OPTION_INPUT,
            "options": sorted([".png", ".exr", ".jpg", ".tif"]),
        },
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        self.set_output("out_str", self.get_attribute_value("internal_str"))


class TextFileExtensionSelect(GeneralLogicNode):
    NICE_NAME = "Text file extension select"

    INTERNALS_DICT = {
        "internal_str": {
            "type": str,
            "gui_type": InputsGUI.OPTION_INPUT,
            "options": sorted([".txt", ".json", ".yml", ".toml", ".csv", ".xml"]),
        },
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        self.set_output("out_str", self.get_attribute_value("internal_str"))


class CodeFileExtensionSelect(GeneralLogicNode):
    NICE_NAME = "Code file extension select"

    INTERNALS_DICT = {
        "internal_str": {
            "type": str,
            "gui_type": InputsGUI.OPTION_INPUT,
            "options": sorted(
                [".py", ".cpp", ".h", ".rs", ".sh", ".ipynb", ".ps1", ".bat"]
            ),
        },
    }

    OUTPUTS_DICT = {"out_str": {"type": str}}

    def run(self):
        self.set_output("out_str", self.get_attribute_value("internal_str"))
